package physbeans.func;

import physbeans.event.*;

/**
 * Describes a function with input and output values.
 * Sends a TriggerEvent to signal general change (f.i. of a parameter) or
 * a VectorEvent if an output vector has been recomputed.
 * <P>Events are sent automatically, as long as autoTriggered is
 * true. Otherwise  trigger() has to be called manually.
 * <p>Children have to implement the abstract compute() method.
 */
abstract public class GenericFunction {
  
  protected boolean autoTriggered;
  
  /**
   * Default constructor.
   */
  public GenericFunction() {
    autoTriggered = true;
    trigSup = new TriggerSupport();
    vecSup = new VectorSupport();
  }
  
  /**
   * Computes output values and informs listeners.<br>
   * This message is used to spread the function result.
   */
  public void trigger() {
    compute();  // compute new output values
    fireVectorEvent(new VectorEvent(this)); // inform listeners
  }
  
  /**
   * Computes output values and informs listeners.<br>
   * This message is used to inform general listeners, not to propagate the
   * function result.
   */
  public void inform() {
    compute();  // compute new output values
    fireTriggerEvent(new TriggerEvent(this));
  }
  
  /**
   * computes new values for output vector.
   */
  abstract protected void compute();
  
  /**
   * Gets value of autoTriggered flag.
   */
  public boolean isAutoTriggered() {
    return this.autoTriggered;
  }
  
  /**
   * Sets value of autoTrigger flag.<br>
   * true -> output values are computed and propagated automatically
   * whenever input values change
   */
  public void setAutoTriggered(boolean autoTriggered) {
    this.autoTriggered = autoTriggered;
  }
  
  // support for Trigger events
  protected transient TriggerSupport trigSup;
  
  /** add a TriggerListener */
  public synchronized void addTriggerListener(TriggerListener l) {
    trigSup.addTriggerListener(l);
    inform();   // inform immediately
  }
  
  /** remove a TriggerListener */
  public synchronized void removeTriggerListener(TriggerListener l) {
    trigSup.removeTriggerListener(l);
  }
  
  /** fire a Trigger event */
  public void fireTriggerEvent(TriggerEvent e) {
    trigSup.fireTriggerEvent(e);
  }
  
  // support for VectorChange events
  protected transient VectorSupport vecSup;
  
  /** register vector listener
   */
  public synchronized void addVectorListener(VectorListener listener) {
    vecSup.addVectorListener(listener);
  }
  /** de-register vector listener
   */
  public synchronized void removeVectorListener(VectorListener listener) {
    vecSup.removeVectorListener(listener);
  }
  /** fire vector event
   */
  public void fireVectorEvent(VectorEvent e) {
    vecSup.fireVectorEvent(e);
  }
}
