package physbeans.func;

import physbeans.math.*;
import physbeans.model.*;
import static java.lang.Math.*;

/**
 * Function that stores a number of data points, given by vectors dataX, dataY
 * of x and y coordinates, and computes a vector of y values from an input
 * vector of x values using linear, polynomial or Spline interpolation.
 */
public class Interpolation1DFunction extends OneParameterFunction {
  
  // possible types of interpolators
  public static final int LINEAR_INTERPOLATOR = 0;
  public static final int POLYNOMIAL_INTERPOLATOR = 1;
  public static final int SPLINE_INTERPOLATOR = 2;
  
  protected DVector dataX, dataY;
  protected int type;
  protected Interpolator1D interpolator;
  
  /**
   * Creates a standard Spline interpolation function.
   */
  public Interpolation1DFunction() {
    super();
    dataX = new DVector(1.0, 2.0, 3.0);
    dataY = new DVector(1.0, 4.0, 9.0);
    type = SPLINE_INTERPOLATOR;
    createInterpolator();
  }
  
  public double evaluate(double t) {
    return interpolator.evaluate(t);
  }
  
  /**
   * Gets the x coordinates of the data points.
   */
  public DVector getDataX() {
    return dataX;
  }
  
  /**
   * Sets the x coordinates of the data points.
   */
  public void setDataX(DVector x) {
    dataX = x;
    createInterpolator();
    inform();
  }
  
  /**
   * Gets the y coordinates of the data points.
   */
  public DVector getDataY() {
    return dataY;
  }
  
  /**
   * Sets the y coordinates of the data points.
   */
  public void setDataY(DVector y) {
    dataY = y;
    createInterpolator();
    inform();
  }
  
  /**
   * Gets type of interpolator.
   * <p>Possible values are
   *  <ul><li>linear interpolator</li>
   *      <li>polynomial interpolator</li>
   *      <li>spline interpolator</li></ul>
   */
  public int getInterpolatorType() {
    return type;
  }
  
  /**
   * Sets type of interpolator.
   * <p>Possible values are
   *  <ul><li>linear interpolator</li>
   *      <li>polynomial interpolator</li>
   *      <li>spline interpolator</li></ul>
   */
  public void setInterpolatorType(int type) {
    this.type = type;
    createInterpolator();
    inform();
  }
  
  /**
   * Creates an interpolator according to the given type.
   */
  protected void createInterpolator() {
    Point2dVector points = new Point2dVector(dataX, dataY);
    switch (type) {
      case  LINEAR_INTERPOLATOR:
        interpolator = new LinearInterpolator(points);
        break;
      case  POLYNOMIAL_INTERPOLATOR:
        interpolator = new NewtonInterpolator(points);
        break;
      case  SPLINE_INTERPOLATOR:
        interpolator = new SplineInterpolator(points);
        break;
      default:
        interpolator = new SplineInterpolator(points);
    }
  }
}
