package physbeans.func;

import physbeans.model.*;
import physbeans.math.*;
import static java.lang.Math.*;

/**
 * Function that computes a number of field values of a given scalar field
 *   along an oblique cut line<br>
 * model: ScalarField   (field to cut through) <br>
 * vector input:  none <br>
 * vector output: <br>
 *     x    coordinate s along the cut line, 0 at center point<br>
 *     y    value of the scalar field at the point corresponding to x
 */
public class ObliqueCutFunction extends GenericVectorFunction {
  
  protected DVector center;
  protected double length, alpha;
  protected Real2DFunction field;
  
  /** Creates a new instance of CutFunction */
  public ObliqueCutFunction() {
    super(0, 2, 100);
    field = null;
    center = new DVector(2);
    alpha = PI/4;
    length = 2*sqrt(2);
  }
  
  /**
   * Computes new values for the output vectors.
   */
  protected void compute() {
    if (field == null) {
      return;
    }
    
    int n = getNPoints();
    
    // start point and increments
    double x = center.x() - cos(alpha)*length/2.0;
    double y = center.y() - sin(alpha)*length/2.0;
    double s = - length/2.0;
    double dx = length*cos(alpha)/(n - 1);
    double dy = length*sin(alpha)/(n - 1);
    double ds = length/(n-1);
    
    for(int i = 0; i < n; i++) {
      outputValues[0].set(i, s);
      outputValues[1].set(i, field.getScalarValue(x, y));
      s += ds;
      x += dx;
      y += dy;
    }
  }
  
  /**
   * Gets angle defining the direction of the cut line.<br>
   * alpha = 0 is the x direction.
   */
  public double getAlpha() {
    return alpha;
  }
  
  /**
   * Sets angle defining the direction of the cut line.<br>
   * alpha = 0 is the x direction.
   */
  public void setAlpha(double al) {
    alpha = al;
    trigger();
  }
  
  /**
   * Gets center point of the cut line.
   */
  public DVector getCenter() {
    return center;
  }
  
  /**
   * Sets center point of the cut line.
   */
  public void setCenter(DVector p) {
    center = p;
    trigger();
  }
  
  /**
   * Gets length of cut line.
   */
  public double getLength() {
    return length;
  }
  
  /**
   * Sets length of cut line.
   */
  public void setLength(double l) {
    length = l;
    trigger();
  }
  
  /**
   * Gets the number of points computed along the cut.
   */
  public int getNPoints() {
    return getOutputLength(0);
  }
  
  /**
   * Sets the number of points computed along the cut.
   */
  public void setNPoints(int nPoints) {
    setAllOutputLengths(nPoints);
  }
  
  /**
   * Returns the values of the independent coordinate (x for a horizontal cut)
   * as a vector.
   */
  public DVector getXVector() {
    return outputValues[0];
  }
  
  /**
   * Returns the values of the dependent coordinate (the scalar field along the
   * cut) as a vector.
   */
  public DVector getYVector() {
    return outputValues[1];
  }
  
  /**
   * Gets the scalar field the CutFunction cuts through.
   */
  public Real2DFunction getScalarField() {
    return field;
  }
  
  /**
   * Sets the scalar field the CutFunction cuts through.
   */
  public void setScalarField(Real2DFunction f) {
    field = f;
    trigger();
  }
}
