package physbeans.func;

import physbeans.math.*;
import physbeans.event.*;
import java.io.*;

/**
 * Generic1dFunction (parameter = time) solving a general ODE.
 */
public class SolveODEFunction extends Generic1dFunction {
  
  protected ODEAdaptiveSolver solver;
  
  /**
   * Standard function, needs setOde afterwards.
   */
  public SolveODEFunction() {
    super(1);
  }
  
  /**
   * Creates standard function for given ode.
   */
  public SolveODEFunction(ODE ode) {
    super(ode.getX0().getDimension());
    solver = new ODEAdaptiveSolver(ode);  // default solver
  }
  
  /**
   * Computes next values for output vector by solving the ODE.
   */
  public DVector computeFunction(double tNext) {
    double dt = tNext - solver.t;
    
    // compute next values for next step
    try {
      solver.nextStep(dt);
    } catch (StepsizeTooSmallException e) {
      e.printStackTrace();
    }
    
    // get current values
    DVector out = solver.x.copy();
    
    return out;
  }
  
  /**
   * Reset function to initial values using the ODE start time.<br>
   * It is used with timer reset or in listeners to parameter changes of the ode.
   */
  public void restart() {
    solver.restart();
    inputValues.set(0, solver.getOde().getT0());
    outputValues = solver.getValues();
  }
  
  /**
   * Resets to initial values but uses the current time.
   */
  public void restartNow() {
    double tNow = inputValues.get(0);
    restart();  // careful, changes current time!
    
    getSolver().setTime(tNow);
    inputValues.set(0, tNow);
  }
  
  /**
   * Sets input time.
   */
  public void setTime(double t) {
    setInputValue(t);
  }
  
  /**
   * Gets input time.<br>
   * Conveniance function.
   */
  protected double getTime() {
    return inputValues.get(0);
  }

  /**
   * Gets ode that is solved by this function.
   */
  public ODE getOde() {
    return solver.getOde();
  }
  
  /**
   * Sets ode that is solved by this function.
   * Sets the output size to the dimension of the ODE.
   */
  public void setOde(ODE ode) {
    setOutputLength(ode.getX0().getDimension());
    solver = new ODEAdaptiveSolver(ode);
  }
  
  /**
   * Gets solver that is used to solve the given ode.
   */
  public ODEAdaptiveSolver getSolver() {
    return this.solver;
  }
  
  /**
   * Sets solver that is used to solve the given ode.
   */
  public void setSolver(ODEAdaptiveSolver solver) {
    this.solver = solver;
  }
  
  /**
   * Sets all initial values.
   */
  public void setInitialValues(DVector x0) {
    getOde().setX0(x0);
    restartNow();                  // set internal values to new start values
  }
  
  /**
   * Gets all initial values.
   */
  public DVector getInitialValues() {
    return getOde().getX0();
  }
  
  /**
   * Sets an initial value.
   */
  public void setInitialValue(int n, double val) {
    getOde().setX0(n, val);
    restartNow();                  // set internal values to new start values
  }
  
  /**
   * Gets an initial value.
   */
  public double getInitialValue(int n) {
    return getOde().getX0().get(n);
  }
  
  /**
   * Sets absolute tolerance of the adaptive solver.
   * <P>Default value: 1e-6
   */
  public void setAbsoluteTolerance(double t) {
    solver.setAbsoluteTolerance(t);
  }
  
  /**
   * Gets absolute tolerance of the adaptive solver.
   */
  public double getAbsoluteTolerance() {
    return solver.getAbsoluteTolerance();
  }
  
/**
   * Gets value of throwStepsizeTooSmall flag<br>
   * If set to true (default) the solver throws an exception when the stepsize
   * gets too small. In rare cases a sudden parameter change can lead to a few
   * very tiny steps, which would block further execution.
   * Set the flag to true there.
   */
  public boolean isThrowStepsizeTooSmall() {
    return solver.isThrowStepsizeTooSmall();
  }

  /**
   * Sets value of throwStepsizeTooSmall flag
   */
  public void setThrowStepsizeTooSmall(boolean flag) {
     solver.setThrowStepsizeTooSmall(flag);
  }
}
