package physbeans.inout;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;

import physbeans.math.*;
import physbeans.event.*;

/**
 * Controls for Screens consisting of a scaling panel and a display panel.<br>
 * The scaling panel allows to change the scale of a screen by sending
 * VectorEvent's with the corresponding scaling factor. Scaling can affect
 * x or y direction or both, according to the changeScale function from Screen
 * that is called. <br>
 * The display panel can show x and y values that originate from a mouse click
 * to a screen and are sent via a VectorEvent.
 */
public class ScreenControls extends PhysicsPanel {
  
  // GUI elements
  protected JLabel scaleTitle;
  protected JButton plusButton; // doubles scale
  protected JLabel scaleLabel; // shows scale
  protected JButton minusButton; // halves scale
  protected NumericTextField xTextField; // display coordinates of marked point
  protected NumericTextField yTextField;
  protected JPanel positionPanel;
  protected JPanel scalePanel;
  
  // strings showing the scale
  // array has to be symmetric around 1 !!
  protected String[] scaleTexts = {"1/16", "1/8", "1/4", "1/2", "1", "2", "4", "8", "16"};
  protected int scaleIndex; // index into scaleTexts
  protected DVector lastScaleChange;   // for simple get function
  
  /** Creates a new instance of Oscilloscope */
  public ScreenControls() {
    super();
    setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
    
    lastScaleChange = new DVector(1.0);
    buildGUI();
  }
  
  /**
   * main part of constructor as extra function to use with other screens
   **/
  protected void buildGUI() {
    scaleTitle = new JLabel("Scale: ", JLabel.CENTER);
    scaleTitle.setFont(null);
    
    // get icons for +/- buttons
    ClassLoader myCL = getClass().getClassLoader();
    ImageIcon minusIcon = null;
    URL minusURL= myCL.getResource("physbeans/resources/minus.gif");
    if (minusURL != null) {
      minusIcon = new ImageIcon(minusURL);
    }
    ImageIcon plusIcon = null;
    URL plusURL = myCL.getResource("physbeans/resources/plus.gif");
    if (plusURL != null) {
      plusIcon = new ImageIcon(plusURL);
    }
    
    plusButton = new JButton(plusIcon);
    plusButton.setOpaque(false);
    plusButton.setMargin(new Insets(1,1,1,1));
    plusButton.addActionListener(new PlusButtonListener());
    
    scaleIndex = scaleTexts.length / 2;    // middle value, == "1"
    scaleLabel = new JLabel(" " + scaleTexts[scaleIndex] + " ", JLabel.CENTER);
    scaleLabel.setFont(null);
    // scaleLabel should not change size
    // size it with standard font
    FontMetrics fm = getFontMetrics(new Font("Dialog", Font.PLAIN, 11));
    int justw = fm.stringWidth("1/16"); // get size of justString using given font
    int justh = fm.getHeight();
    scaleLabel.setPreferredSize(new Dimension(justw + 4, justh));
    
    minusButton = new JButton(minusIcon);
    minusButton.setOpaque(false);
    minusButton.setMargin(new Insets(1,1,1,1));
    minusButton.addActionListener(new MinusButtonListener());
    
    xTextField = new NumericTextField(0.0, 5, 3, "Position:", "");
    xTextField.setEditable(false);
    yTextField = new NumericTextField(0.0, 5, 3, "", "");
    yTextField.setEditable(false);
    
    //  build GUI
    // all scale things into scalePanel
    scalePanel = new JPanel(new FlowLayout(FlowLayout.CENTER, 2, 3));
    scalePanel.setOpaque(false);
    scalePanel.setFont(null);
    scalePanel.add(scaleTitle);
    scalePanel.add(plusButton);
    scalePanel.add(scaleLabel);
    scalePanel.add(minusButton);
    
    // displays for click position into positionPanel
    positionPanel = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 3));
    positionPanel.setOpaque(false);
    positionPanel.setFont(null);
    positionPanel.add(xTextField);
    positionPanel.add(yTextField);
    
    // controlPanel combines all controls
    add(scalePanel);
    add(Box.createHorizontalGlue());
    add(positionPanel);
  }
  
  /**
   * Sets label of scale controls.
   */
  public void setScaleLabel(String s) {
    scaleTitle.setText(s);
  }
  
  /**
   * Gets label of scale controls.
   */
  public String getScaleLabel() {
    return scaleTitle.getText();
  }
  
  /**
   * Sets label of position text fields.
   */
  public void setPositionLabel(String s) {
    xTextField.setDescription(s);
  }
  
  /**
   * Gets label of position text fields.
   */
  public String getPositionLabel() {
    return xTextField.getDescription();
  }
  
  /**
   * divide scale by a factor of two
   */
  protected void decreaseScale() {
    // do nothing if minimal value reached
    if (scaleIndex == 0) {
      return;
    }
    
    // change label
    scaleIndex--;
    scaleLabel.setText(scaleTexts[scaleIndex]);
    changeScale(2.0);
  }
  
  /**
   * enlarge scale by a factor of two
   */
  protected void increaseScale() {
    // do nothing if maximal value reached
    if (scaleIndex == scaleTexts.length - 1) {
      return;
    }
    
    // ok it is not maximal
    scaleIndex++;
    scaleLabel.setText(scaleTexts[scaleIndex]);
    changeScale(0.5);
  }
  
  /**
   * Changes scale of the screen.
   */
  protected void changeScale(double d) {
    lastScaleChange = new DVector(d);
    fireVectorEvent(new VectorEvent(this, lastScaleChange));
  }
  
  /**
   * Sets initial value of the scale index.
   */
  public void setInitialScaleIndex(int newIndex) {
    while (scaleIndex < newIndex) {
      increaseScale();
    }
    
    while (scaleIndex > newIndex) {
      decreaseScale();
    }
  }
  
  /**
   * Gets initial value of the scale index.
   */
  public int getInitialScaleIndex() {
    return scaleIndex;
  }
  
  /**
   * Gets last scale change.
   * Easy access function for listeners.
   */
  public DVector getScaleChange() {
    return lastScaleChange;
  }
  
  /**
   * Sets text fields according to x and y coordinate of cector v.
   */
  public void setTextFields(DVector v) {
    xTextField.setValue(v.get(0));
    yTextField.setValue(v.get(1));
  }
  
  // listeners
  protected class PlusButtonListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
      increaseScale();
    }
  }
  protected class MinusButtonListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
      decreaseScale();
    }
  }
  
  // the usual support for VectorChange events
  
  protected transient VectorSupport vecSup = new VectorSupport();
  
  /** register vector listener
   */
  public synchronized void addVectorListener(VectorListener listener) {
    vecSup.addVectorListener(listener);
  }
  /** de-register vector listener
   */
  public synchronized void removeVectorListener(VectorListener listener) {
    vecSup.removeVectorListener(listener);
  }
  /** fires vector event
   */
  public void fireVectorEvent(VectorEvent e) {
    vecSup.fireVectorEvent(e);
  }
}

