package physbeans.math;

import physbeans.model.*;

/**
 * Matrix of complex numbers.
 */
public class CMatrix {

  protected double[][] re,  im;
  protected int m,  n;

  // implementation note: it seems more natural to use an array of Complex,
  // but that would have a lousy performance!
  /**
   * Construct an m-by-n matrix of zeros.
   */
  public CMatrix(int m, int n) {
    re = new double[m][n];
    im = new double[m][n];
    this.m = m;
    this.n = n;
  }

  /**
   * Construct a complex matrix from given matrices for the real and
   * imaginary values.
   */
  public CMatrix(Matrix re, Matrix im) {
    this.re = re.getArray();
    this.im = im.getArray();
    m = this.re.length;
    n = this.re[0].length;
  }

  /**
   * Construct a matrix from 2d-arrays for real and imaginary parts.
   * Arrays are just stored, not copied.
   */
  public CMatrix(double[][] a, double[][] b) {
    re = a;
    im = b;
    m = this.re.length;
    n = this.re[0].length;
  }

  /**
   * Construct a matrix from 2d-array for real parts, imaginary parts are 0.
   */
  public CMatrix(double[][] a) {
    re = a;
    m = this.re.length;
    n = this.re[0].length;
    im = new double[m][n];
  }

  /**
   * Construct an m-by-n constant matrix.
   */
  public CMatrix(int m, int n, Complex c) {
    re = new double[m][n];
    im = new double[m][n];
    this.m = m;
    this.n = n;

    double cRe = c.real();
    double cIm = c.imag();
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        re[i][j] = cRe;
        im[i][j] = cIm;
      }
    }
  }

  /**
   * Construct a matrix from two one-dimensional packed arrays defining
   * real and imaginary parts.
   *
   * @param a real part as one-dimensional array of doubles, packed by columns.
   * @param b imaginary part as one-dimensional array of doubles, packed by columns.
   * @param m number of rows.
   */
  public CMatrix(double[] a, double[] b, int m) {
    this.m = m;
    n = (m != 0 ? a.length / m : 0);
    if (m * n != a.length) {
      throw new IllegalArgumentException("Array length must be a multiple of m.");
    }

    re = new double[m][n];
    im = new double[m][n];
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        re[i][j] = a[i + j * m];
        im[i][j] = b[i + j * m];
      }
    }
  }

  /**
   * Gets column (second) dimension.
   */
  public int getColumnDimension() {
    return n;
  }

  /**
   * Gets row (first) dimension.
   */
  public int getRowDimension() {
    return m;
  }

  /**
   * Sets a single element.
   */
  public void set(int i, int j, Complex c) {
    re[i][j] = c.real();
    im[i][j] = c.imag();
  }

  /**
   * Sets a single element with a real value.
   */
  public void set(int i, int j, double r) {
    re[i][j] = r;
    im[i][j] = 0.0;
  }

  /**
   * Gets a single element.
   */
  public Complex get(int i, int j) {
    return new Complex(re[i][j], im[i][j]);
  }

  /**
   * Gets the column vector of given index col.
   */
  public CVector getColumnVector(int col) {
    DVector vRe = new DVector(m);
    DVector vIm = new DVector(m);
    for (int i = 0; i < m; i++) {
      vRe.set(i, re[i][col]);
      vIm.set(i, im[i][col]);
    }

    return new CVector(vRe, vIm);
  }

  /**
   * Sets the column vector of given index col.
   */
  public void setColumnVector(int col, CVector colVec) {
    double[] vRe = colVec.getReal().getArray();
    double[] vIm = colVec.getImag().getArray();
    for (int i = 0; i < m; i++) {
      re[i][col] = vRe[i];
      im[i][col] = vIm[i];
    }
  }

  /**
   * Gets the row vector of given index row.
   */
  public CVector getRowVector(int row) {
    DVector vRe = new DVector(n);
    DVector vIm = new DVector(n);
    for (int i = 0; i < n; i++) {
      vRe.set(i, re[row][i]);
      vIm.set(i, im[row][i]);
    }

    return new CVector(vRe, vIm);
  }

  /**
   * Sets the row vector of given index row.
   */
  public void setRowVector(int row, CVector rowVec) {
    double[] vRe = rowVec.getReal().getArray();
    double[] vIm = rowVec.getImag().getArray();
    for (int i = 0; i < n; i++) {
      re[row][i] = vRe[i];
      im[row][i] = vIm[i];
    }
  }

  /**
   * Frobenius norm
   *
   * @return sqrt of sum of squares of all elements.
   */
  public double normF() {
    // slow, too many sqrt
    // TODO: rewrite with max etc.

    double f = 0;
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        f = Math.hypot(f, re[i][j]);
        f = Math.hypot(f, im[i][j]);
      }

    }
    return f;
  }

  /**
   * Add a matrix to another one, C = A + B
   */
  public CMatrix plus(CMatrix B) {
    CMatrix X = new CMatrix(m, n);
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        X.re[i][j] = re[i][j] + B.re[i][j];
        X.im[i][j] = im[i][j] + B.im[i][j];
      }
    }
    return X;
  }

  /**
   * Add a matrix to another one in place, A = A + B
   */
  public CMatrix plusEquals(CMatrix B) {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        re[i][j] += B.re[i][j];
        im[i][j] += B.im[i][j];
      }
    }
    return this;
  }

  /**
   * Multiplies matrix to another one, C = A * B
   */
  public CMatrix times(CMatrix B) {
    if (B.m != n) {
      throw new IllegalArgumentException("Matrix inner dimensions must agree.");
    }

    CMatrix X = new CMatrix(m, B.n);
    for (int i = 0; i < m; i++) {
      for (int k = 0; k < B.n; k++) {
        double s = 0;
        double t = 0;
        for (int j = 0; j < n; j++) {
          s += re[i][j] * B.re[j][k] - im[i][j] * B.im[j][k];
          t += re[i][j] * B.im[j][k] + im[i][j] * B.re[j][k];
        }
        X.re[i][k] = s;
        X.im[i][k] = t;
      }
    }

    return X;
  }

  /**
   * Multiply a matrix by a real scalar, C = s*A
   */
  public CMatrix times(double s) {
    CMatrix X = new CMatrix(m, n);
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        X.re[i][j] = s * re[i][j];
        X.im[i][j] = s * im[i][j];
      }
    }
    return X;
  }

  /**
   * Multiply a matrix by a real scalar in place, A = s*A
   */
  public CMatrix timesEquals(double s) {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        re[i][j] = s * re[i][j];
        im[i][j] = s * im[i][j];
      }
    }
    return this;
  }

  /**
   * Multiply a matrix by a complex scalar, V = s*A
   */
  public CMatrix times(Complex c) {
    CMatrix X = new CMatrix(m, n);
    double s = c.real();
    double t = c.imag();
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        X.re[i][j] = s * re[i][j] - t * im[i][j];
        X.im[i][j] = t * re[i][j] + s * im[i][j];
      }
    }
    return X;
  }

  /**
   * Multiply a matrix by a complex scalar in place, A = c*A
   */
  public CMatrix timesEquals(Complex c) {
    double s = c.real();
    double t = c.imag();
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        double reTemp = s * re[i][j] - t * im[i][j];
        im[i][j] = t * re[i][j] + s * im[i][j];
        re[i][j] = reTemp;
      }
    }
    return this;
  }

  /**
   * Multiplies matrix pointwise to another one, C = A .* B
   */
  public CMatrix arrayTimes(CMatrix B) {
    CMatrix X = new CMatrix(m, n);
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        X.re[i][j] = re[i][j] * B.re[i][j] - im[i][j] * B.im[i][j];
        X.im[i][j] = re[i][j] * B.im[i][j] + im[i][j] * B.re[i][j];
      }
    }
    return X;
  }

  /**
   * Multiply a matrix pointwise to another one in place, A = A .* B
   */
  public CMatrix arrayTimesEquals(CMatrix B) {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        double reTemp = re[i][j] * B.re[i][j] - im[i][j] * B.im[i][j];
        im[i][j] = re[i][j] * B.im[i][j] + im[i][j] * B.re[i][j];
        re[i][j] = reTemp;
      }
    }
    return this;
  }

  /**
   * Divides matrix pointwise by another one, C = A ./ B
   */
  public CMatrix arrayRightDivide(CMatrix B) {
    CMatrix X = new CMatrix(m, n);
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        double fac = 1.0 / (B.re[i][j] * B.re[i][j] + B.im[i][j] * B.im[i][j]);
        X.re[i][j] = (re[i][j] * B.re[i][j] + im[i][j] * B.im[i][j]) * fac;
        X.im[i][j] = (-re[i][j] * B.im[i][j] + im[i][j] * B.re[i][j]) * fac;
      }
    }
    return X;
  }

  /**
   * Divides a matrix pointwise by another one in place, A = A ./ B
   */
  public CMatrix arrayRightDivideEquals(CMatrix B) {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        double fac = 1.0 / (B.re[i][j] * B.re[i][j] + B.im[i][j] * B.im[i][j]);
        double reTemp = (re[i][j] * B.re[i][j] + im[i][j] * B.im[i][j]) * fac;
        im[i][j] = (-re[i][j] * B.im[i][j] + im[i][j] * B.re[i][j]) * fac;
        re[i][j] = reTemp;
      }
    }
    return this;
  }

  /**
   * Returns matrix with elements that are shifted by p elements vertically
   * and q elements horizontally, filling with 0's.
   * Positive shift means down/right.
   */
  public CMatrix shift(int p, int q) {
    CMatrix B = new CMatrix(m, n);
    int i0In, i1In, j0In, j1In, i0Out, i1Out, j0Out, j1Out;

    // first index / vertical
    if ((p >= 0) & (p < m)) {           // down shift
      i0In = 0;
      i1In = m - p - 1;
      i0Out = p;
      i1Out = m - 1;
    } else if ((p < 0) & (p > -m)) {    // up shift
      i0In = -p;
      i1In = m - 1;
      i0Out = 0;
      i1Out = m + p - 1;
    } else {                            // out shift
      return B;                         // all 0's
    }

    // second index / horizontal
    if ((q >= 0) & (q < n)) {           // right shift
      j0In = 0;
      j1In = n - q - 1;
      j0Out = q;
      j1Out = n - 1;
    } else if ((q < 0) & (q > -n)) {    // left shift
      j0In = -q;
      j1In = n - 1;
      j0Out = 0;
      j1Out = n + q - 1;
    } else {                            // out shift
      return B;                         // all 0's
    }

    for (int i = i0Out; i <= i1Out; i++) {
      for (int j = j0Out; j <= j1Out; j++) {
        B.re[i][j] = re[i + i0In - i0Out][j + j0In - j0Out];
        B.im[i][j] = im[i + i0In - i0Out][j + j0In - j0Out];
      }

    }
    return B;
  }

  /**
   * Pointwise square root of a matrix, A = sqrt(A)
   * with branch cut along the negative real axis.
   */
  public CMatrix sqrt() {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        Complex z = Complex.sqrt(new Complex(re[i][j], im[i][j]));
        re[i][j] = z.real();
        im[i][j] = z.imag();
      }
    }
    return this;
  }

  /**
   * Pointwise exp of a matrix, A = exp(A)
   */
  public CMatrix exp() {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        Complex z = Complex.exp(new Complex(re[i][j], im[i][j]));
        re[i][j] = z.real();
        im[i][j] = z.imag();
      }
    }
    return this;
  }

  /**
   * Pointwise cosine of a matrix, A = cos(A)
   */
  public CMatrix cos() {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        Complex z = Complex.cos(new Complex(re[i][j], im[i][j]));
        re[i][j] = z.real();
        im[i][j] = z.imag();
      }
    }
    return this;
  }

  /**
   * Pointwise sine of a matrix, C = sin(A)
   */
  public CMatrix sin() {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        Complex z = Complex.sin(new Complex(re[i][j], im[i][j]));
        re[i][j] = z.real();
        im[i][j] = z.imag();
      }
    }
    return this;
  }

  /** 
   * Gets a submatrix given by two IndexRange's.
   *
   * @param idx
   * @param idy
   * @return A(idx,idy)
   */
  public CMatrix getMatrix(IndexRange idx, IndexRange idy) {
    CMatrix B = new CMatrix(idx.getSize(), idy.getSize());
    int i, j, i1, j1;
    if (idx.getStride() > 0) {
      for (i = idx.getStart(), i1 = 0; i <= idx.getEnd(); i += idx.getStride(), i1++) {
        if (idy.getStride() > 0) {
          for (j = idy.getStart(), j1 = 0; j <= idy.getEnd(); j += idy.getStride(), j1++) {
            B.re[i1][j1] = re[i][j];
            B.im[i1][j1] = im[i][j];
          }
        } else {
          for (j = idy.getStart(), j1 = 0; j >= idy.getEnd(); j += idy.getStride(), j1++) {
            B.re[i1][j1] = re[i][j];
            B.im[i1][j1] = im[i][j];
          }
        }
      }
    } else {
      for (i = idx.getStart(), i1 = 0; i >= idx.getEnd(); i += idx.getStride(), i1++) {
        if (idy.getStride() > 0) {
          for (j = idy.getStart(), j1 = 0; j <= idy.getEnd(); j += idy.getStride(), j1++) {
            B.re[i1][j1] = re[i][j];
            B.im[i1][j1] = im[i][j];
          }
        } else {
          for (j = idy.getStart(), j1 = 0; j >= idy.getEnd(); j += idy.getStride(), j1++) {
            B.re[i1][j1] = re[i][j];
            B.im[i1][j1] = im[i][j];
          }
        }
      }
    }
    return B;
  }

  /**
   * Sets part of the (large) this matrix from a (small) X matrix.
   */
  public void setSubMatrix(IndexRange idx, IndexRange idy, FastCMatrix X) {
    // X stores by rows, reading is in row order
    int i, j, i1, j1;
    int idx0 = idx.getStart();
    int idxs = idx.getStride();
    int idy0 = idy.getStart();
    int idys = idy.getStride();

    for (i1 = 0; i1 < X.m; i1++) {
      i = idx0 + i1 * idxs;
      for (j1 = 0; j1 < X.n; j1++) {
        j = idy0 + j1 * idys;
        re[i][j] = X.re[j1 + X.n * i1];
        im[i][j] = X.im[j1 + X.n * i1];
      }
    }
  }

  /**
   * Gets the real part of the matrix.
   */
  public double[][] getRealArray() {
    return re;
  }

  /**
   * Gets the imaginary part of the matrix.
   */
  public double[][] getImagArray() {
    return im;
  }

  /**
   * Makes a deep copy of a matrix.
   */
  public CMatrix copy() {
    CMatrix X = new CMatrix(m, n);
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        X.re[i][j] = re[i][j];
        X.im[i][j] = im[i][j];
      }
    }
    return X;
  }

  /**
   * Sets the matrix to X (A = X) by copying the values.
   */
  public void setMatrix(CMatrix X) {
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        re[i][j] = X.re[i][j];
        im[i][j] = X.im[i][j];
      }
    }
  }
}
