package physbeans.math;

/**
 * class containing fft related functions.
 */
public class FFT {
  
  /**
   * Computes fourier transform of complex vector z.
   * Simple slow method: uses direct evaluation of the sum.
   * Works with any size of input vector z.
   */
  public static CVector simpleTransform(CVector z) {
    int N = z.getDimension();
    CVector X = new CVector(N);
    
    for (int k = 0; k <= N-1; k++) {
      double fac = -2.0*Math.PI*k/N;
      for (int n = 0; n <= N-1; n++) {
        Complex z1 = Complex.exp(new Complex(0.0, n*fac));
        Complex z2 = z1.times(z.get(n));
        Complex z3 = z2.plus(X.get(k));
        X.set(k, z3);
      }
    }
    return X;
  }
  
  /**
   * Computes direct fft of complex vector z.
   * Assumes, length of input vector z is a power of 2.
   */
  public static CVector transform(CVector z) {
    double[] re = z.getReal().getArray();
    double[] im = z.getImag().getArray();
    int n = z.getDimension();
    
    // check whether n is a power of 2
    double nlog = (int)( Math.log( (double)n )/Math.log(2) + 0.5 );
    if ( (int)Math.round(Math.pow(2, nlog)) != n ) {
      throw new IllegalArgumentException("vector length has to be a power of 2!");
    }
    
    // bit reversal
    int nhalf = n/2;
    int j = 1;
    double temp;
    for (int i = 1; i < n; i++) {
      if (i < j) {
        temp = re[i - 1];
        re[i - 1] = re[j - 1];
        re[j - 1] = temp;
        
        temp = im[i - 1];
        im[i - 1] = im[j - 1];
        im[j - 1] = temp;
      }
      int k = nhalf;
      while (k < j) {
        j = j - k;
        k = k / 2;
      }
      j = j + k;
    }
    
    // after Danielson-Lanczos 
    for (int l = 1; l <= nlog; l++) {
      int lpow2 = (int)(Math.exp(l * Math.log(2)) + 0.5);
      int lpow2half = lpow2 / 2;
      double ur = 1.0;
      double ui = 0.0;
      double wr =  Math.cos(Math.PI/lpow2half);
      double wi = -Math.sin(Math.PI/lpow2half);
      double tempr, tempi;
      for (j = 1; j <= lpow2half; j++) {
        for (int i = j; i <= n; i += lpow2) {
          int ip = i + lpow2half;
          tempr = re[ip - 1] * ur - ui * im[ip - 1];
          tempi = im[ip - 1] * ur + ui * re[ip - 1];
          
          re[ip - 1] = re[i - 1] - tempr;
          im[ip - 1] = im[i - 1] - tempi;
          
          re[i - 1] =  re[i - 1] + tempr;
          im[i - 1] =  im[i - 1] + tempi;
        }
        temp = ur * wr - wi * ui;
        ui = wr * ui + wi * ur;
        ur = temp;
      }
    }
    
    return new CVector(re, im);
  }
  
  /**
   * Computes getReal coefficients of cosine series from complex fourier transform.
   */
  public static DVector getCosineCoeffs(CVector X) {
    // implements a(l) = c(l) + c(N-l)  for l = 0, .. N/2
    // where c(l) = X(l)/N
    // a(l) is getReal, assuming X is computed from a getReal function
    
    int N = X.getDimension();
    int lMax = N/2 + 1;
    DVector a = new DVector(lMax);
    
    // first index is special since X(N) does not exist.
    a.set(0, 2.0*X.get(0).real());
    
    for (int l=1; l < lMax; l++) {
      a.set(l, X.get(l).real() + X.get(N - l).real());
    }
    
    // change scaling
    a.timesEquals(1.0/N);
    return a;
  }
  
  /**
   * Computes getReal coefficients of sine series from complex fourier transform.
   */
  public static DVector getSineCoeffs(CVector X) {
    // implements b(l) = i*(c(l) - c(N-l))  for l = 0, .. N/2
    // where c(l) = X(l)/N
    // b(l) is getReal, assuming X is computed from a getReal function
    
    int N = X.getDimension();
    int lMax = N/2 + 1;
    DVector b = new DVector(lMax);
    
    // first index is 0.
    b.set(0, 0.0);
    
    for (int l=1; l < lMax; l++) {
      b.set(l, -X.get(l).imag() + X.get(N - l).imag());
    }
    
    // change scaling
    b.timesEquals(1.0/N);
    return b;
  }
  
  /**
   * Computes getReal amplitudes from complex fourier transform.
   * They are given as C(l) from a harmonic decomposition
   *  <code>  x(t) = sum C(l) cos(omega l t + psi(l)) </code>
   * where X is the fft of x.
   */
  public static DVector getSpectralAmplitudes(CVector X) {
    DVector a = getCosineCoeffs(X);
    DVector b = getSineCoeffs(X);
    int n = a.getDimension();
    
    DVector C = new DVector(n);
    for (int i = 0; i < n; i++) {
      C.set(i, Math.sqrt(a.get(i)*a.get(i) + b.get(i)*b.get(i)));
    }
    return C;
  }
  
  /**
   * Computes getReal getPhase shifts from complex fourier transform.
   * They are given as psi(l) from a harmonic decomposition
   *  <code>  x(t) = sum C(l) cos(omega l t + psi(l)) </code>
   * where X is the fft of x.
   */
  public static DVector getSpectralPhases(CVector X) {
    DVector a = getCosineCoeffs(X);
    DVector b = getSineCoeffs(X);
    int n = a.getDimension();
    
    DVector psi = new DVector(n);
    for (int i = 0; i < n; i++) {
      psi.set(i, Math.atan2(-b.get(i), a.get(i)));
    }
    return psi;
  }
}
