package physbeans.math;

import static java.lang.Math.*;

/**
 * contains additional mathematical functions
 */
public class Maths {

  public static String infinityString = "infinity";
  protected static double eps = 1e-8;   // accuracy of double comparison

  /**
   * Evaluates the series of Chebyshev polynomials Ti at argument x/2.
   * The series is given by
   * <pre>
   *        N-1
   *         - '
   *  y  =   >   coef[i] T (x/2)
   *         -            i
   *        i=0
   * </pre>
   * Coefficients are stored in reverse order, i.e. the zero
   * order term is last in the array.  Note N is the number of
   * coefficients, not the order.
   * <p>
   * If coefficients are for the interval a to b, x must
   * have been transformed to x -> 2(2x - b - a)/(b-a) before
   * entering the routine.  This maps x from (a, b) to (-1, 1),
   * over which the Chebyshev polynomials are defined.
   * <p>
   * If the coefficients are for the inverted interval, in
   * which (a, b) is mapped to (1/b, 1/a), the transformation
   * required is x -> 2(2ab/x - b - a)/(b-a).  If b is infinity,
   * this becomes x -> 4a/x - 1.
   * <p>
   * SPEED:
   * <p>
   * Taking advantage of the recurrence properties of the
   * Chebyshev polynomials, the routine requires one more
   * addition per loop than evaluating a nested polynomial of
   * the same degree.
   *
   * @param x argument to the polynomial.
   * @param coef the coefficients of the polynomial.
   */
  static double csevl(double x, double coef[]) {
    double b0, b1, b2, twox;
    int i;
    b1 = 0.0;
    b0 = 0.0;
    b2 = 0.0;
    twox = 2.0 * x;
    for (i = coef.length - 1; i >= 0; i--) {
      b2 = b1;
      b1 = b0;
      b0 = twox * b1 - b2 + coef[i];
    }
    return 0.5 * (b0 - b2);
  }

  /**
   * Simple signum function.
   */
  public static double sign(double x) {
    if (x == 0) {
      return 0;
    } else if (x > 0) {
      return 1;
    } else {
      return -1;
    }
  }

  /**
   * Simple square function .
   */
  public static double sqr(double x) {
    return x * x;
  }

  /**
   * Checks, whether two doubles are equal (within a given relative accuracy).
   */
  public static boolean isEqual(double d1, double d2) {
    return (Math.abs(d1 - d2) / Math.max(Math.abs(d1), Math.abs(d2)) < eps);
  }

  /**
   * Checks, whether a double is strictly positive (i.e. larger than 0 including
   * a given accuracy).
   */
  public static boolean isStrictlyPositive(double d) {
    return (d > eps);
  }

  /**
   * Checks, whether a double is zero (within a given absolute accuracy).
   */
  public static boolean isZero(double d) {
    return (Math.abs(d) < eps);
  }

  /**
   * Returns value of accuracy used for comparison of doubles.
   */
  public static double getAccuracy() {
    return eps;
  }

  /**
   * Sets value of accuracy used for comparison of doubles.
   */
  public static void setAccuracy(double eps) {
    Maths.eps = eps;
  }

  /**
   * returns String representing the given double value f. <br>
   * formats according to given number of significant digits prec.<br>
   * uses decimal notation for 1e-2 &lt;= f &lt; 1e3, scientific notation otherwise.
   * If precision = -1, the integer part is shown.
   */
  public static String toString(double f, int prec) {
    // basic problems:
    //    - can't rely on log and multiplication/division by 10
    //    - internal toString uses decimal notation up to 1e7

    if (Double.isNaN(f)) {
      return "NaN";
    }

    int maxDec = 2;   // max no. of digits before using E notation
    int minDec = -2;  // -max no. of zeroes before using E notation

    String decimalPoint = ".";    // no localization used
    String expSymbol = "E";

    // care for sign
    String result = "";
    if (f < 0.0) {
      result = "-";
      f = -f;
    }

    if (prec > 0) {
      // get decimal exponent
      long log10 = Math.round(Math.log(f) / Math.log(10.0) - 0.5);
      if (log10 < -305) {
        return "0.0";
      }
      if (log10 > 307) {
        return infinityString;
      }

      // get digits with 10^(prec-1) <= digits < 10^prec
      double coef = Math.pow(10, prec - log10 - 1);
      long digits = Math.round(f * coef);

      // rounding could change the log10 value of f (f.i. 0.999 -> 1.000),
      //  so log10 and digits have to be corrected
      f = digits / coef;                 // rounded value of f
      log10 = Math.round(Math.log(f) / Math.log(10.0) - 0.5);
      coef = Math.pow(10, prec - log10 - 1);
      digits = Math.round(f * coef);

      if (log10 > maxDec || log10 < minDec) {
        // use scientific notation
        long div = Math.round(Math.pow(10, prec - 1));
        result += (digits / div) + decimalPoint;

        // digits after the decimal point have to be filled with leading 0's
        String digitsAfterPoint = "" + (digits % div);
        while (digitsAfterPoint.length() < prec - 1) {
          digitsAfterPoint = "0" + digitsAfterPoint;
        }
        result += digitsAfterPoint;

        // append exponent
        result += expSymbol + log10;
      } else if (log10 < 0) {
        // use decimal notation 0.xxx
        result += "0" + decimalPoint;
        for (int i = 0; i < -log10 - 1; i++) {
          result += "0";
        }
        result += digits;
      } else {
        // use decimal notation dig.its

        // add zeros, if prec < 1 + log10, f.i. 973.2 with prec = 2
        // this leads to an increase in prec!
        long nZeroes = 1 + log10 - prec;
        for (int i = 0; i < nZeroes; i++) {
          digits *= 10;
          prec++;
        }

        // digits before decimal point
        long div = Math.round(Math.pow(10, prec - log10 - 1));
        result += (digits / div) + decimalPoint;

        // digits after decimal point
        long rem = digits % div;

        // may need additional zeroes in front, f.i. 3.010
        if (rem == 0) {
          //just show zeroes
          for (int i = 0; i < prec - log10 - 1; i++) {
            result += "0";
          }
        } else {
          for (rem *= 10; rem < div; rem *= 10) {
            result += "0";
          }
          result += (digits % div);
        }
      }
    } else {
      result += Long.toString(Math.round(f));
    }
    return result;
  }

  /*
   * Converts double to String for Java initialization.<br>
   * Differs from Double.toString by the handling of infinity.
   */
  static public String toJavaString(double d) {
    if (d == Double.POSITIVE_INFINITY) {
      return "Double.POSITIVE_INFINITY";
    } else if (d == Double.NEGATIVE_INFINITY) {
      return "Double.NEGATIVE_INFINITY";
    } else {
      return Double.toString(d);
    }
  }
}
