package physbeans.math;

import static java.lang.Math.*;     // readable math at last!
import static physbeans.math.Maths.*;

/**
 * Twodimensional horizontal parabola, given by its vertex and the curvature
 * radius at the vertex.
 * It is given by the equation<br>
 *    x - cx = (y - cy)<sup>2</sup>/(2r)<br>
 * Optionally segment boundaries can be given to define a vertical segment of
 * the parabola.
 */
public class Parabola2D {
  
  protected DVector c;
  protected double r;
  protected double xl, xr;
  
  /**
   * Creates a new parabola with vertex c and curvature radius r at the vertex.
   * The segment bounds lie outside the parabola.
   */
  public Parabola2D(DVector c, double r) {
    this.c = c;
    this.r = r;
    
    xl = (r > 0) ? c.x() : c.x() + r;
    xr = (r > 0) ? c.x() + r : c.x();
  }
  
  /**
   * Creates a new parabola with vertex c, curvature radius r at the vertex
   * and given segment bounds.
   */
  public Parabola2D(DVector c, double r, double xl, double xr) {
    this(c, r);
    setSegmentBounds(xl, xr);
  }
  
  /**
   * Sets the boundaries of a vertical parabola segment.
   * It is defined by the x coordinates of left and right edges.
   */
  public void setSegmentBounds(double xl, double xr) {
    this.xl = xl;
    this.xr = xr;
  }
  
  /**
   * Computes an outward pointing unit normal vector of the parabola at a point p.
   * Assumes that the point lies on the parabola.
   */
  public DVector getNormal(DVector p) {
    double px = p.x() - c.x();
    double py = p.y() - c.y();
    
    double r = py*py/(2*px);
    DVector n = new DVector(-r, py);
    n.timesEquals(1.0/n.norm());
    return n;
  }
  
  /**
   * Computes the line parameters of the hit points of a line and the parabola.
   * Returns NaN vector, if no hit point exists.
   */
  protected DVector getRayHitParameterAux(Ray2D ray) {
    
    // solve corresponding quadratic equation for line parameter l
    double sa = sin(ray.getAngle());
    double ca = cos(ray.getAngle());
    DVector p = ray.getStartPoint();
    
    if (isZero(sa)) {
      double l = (sqr(p.y() - c.y()) - 2*r*(p.x() - c.x()))/(2*r*ca);
      return new DVector(l, Double.NaN);
    } else {
      
      double a = (p.y() - c.y())/sa - r*ca/(sa*sa);
      double b = (sqr(p.y() - c.y()) - 2*r*(p.x() - c.x()))/(sa*sa);
      
      double discri = a*a - b;
      if (discri < 0) {
        return new DVector(Double.NaN, Double.NaN);
      }
      double l1 = -a - sqrt(discri);
      double l2 = -a + sqrt(discri);
      return new DVector(l1, l2);
    }
  }
  
  /**
   * Computes the ray parameter of the first hit point of a ray and the parabola.
   * Returns NaN, if no hit point exists.
   */
  public double getRayHitParameter(Ray2D ray) {
    
    DVector l = getRayHitParameterAux(ray);
    
    if (isStrictlyPositive(l.get(0))) {  // smaller parameter gives hit, if positive
      return l.get(0);
    }
    if (isStrictlyPositive(l.get(1))) {  // ok, now let's try the other parameter
      return l.get(1);
    }
    
    return Double.NaN;   // ok, no hits
  }
  
  /**
   * Computes the first hit point of a ray and the parabola. Returns vector of NaN's,
   * if no hit point exists.
   */
  public DVector getRayHitPoint(Ray2D ray) {
    
    double l = getRayHitParameter(ray);
    if (Double.isNaN(l)) {
      return new DVector(Double.NaN, Double.NaN);
    } else {
      return ray.getPointAtParameter(l);
    }
  }
  
  /**
   * Computes the ray parameter of the first hit point of a ray and a vertical
   * parabola segment. Returns NaN, if no hit point exists.
   * The segment is defined by the x coordinates of left and right edges.
   */
  public double getSegmentRayHitParameter(Ray2D ray) {
    
    DVector l = getRayHitParameterAux(ray);
    double l1 = l.get(0);
    double l2 = l.get(1);
    
    double px = ray.getStartPoint().x();
    double ca = cos(ray.getAngle());
    
    if (isStrictlyPositive(l1)) {  // try smaller parameter first, if positive
      double xh = px + l1*ca;
      if (xh >= xl && xh <= xr) {
        return l1;
      }
    }
    if (isStrictlyPositive(l2)) {  // ok, now let's try the other parameter
      double xh = px + l2*ca;
      if (xh >= xl && xh <= xr) {
        return l2;
      }
    }
    
    return Double.NaN;   // ok, no hits
  }
  
  /**
   * Computes the first hit point of a ray and a vertical parabola segment.
   * Returns vector of NaN's, if no hit point exists. <p>
   */
  public DVector getSegmentRayHitPoint(Ray2D ray) {
    
    double l = getSegmentRayHitParameter(ray);
    if (Double.isNaN(l)) {
      return new DVector(Double.NaN, Double.NaN);
    } else {
      return ray.getPointAtParameter(l);
    }
  }
}
