package physbeans.model;


import physbeans.phys.*;
import physbeans.math.*;
import physbeans.event.*;

/*
 * Base class of scalar fields in n space dimensions.<P>
 * Informs listeners of global changes with a trigger event.<br>
 * Can send values at a single position as vector event.
 */
public abstract class ScalarField extends PhysModel implements Real2DFunction {
  
  protected int dim;
  protected DVector lastPosition;
  
  /**
   * default scalar field: 2 dimensions
   */
  public ScalarField() {
    this(2);
  }
  
  /**
   * scalar field with dim dimensions
   */
  public ScalarField(int dim) {
    super();
    this.dim = dim;
    lastPosition = new DVector(dim);
    vecSup = new VectorSupport();
  }
  
  /**
   * return the value at the given coordinates
   */
  abstract public double getScalarValue(DVector coord);
  
  /**
   * return the value at the given coordinates<br>
   * special variant for the (common) 2d case
   */
  public double getScalarValue(double x, double y) {
    // default case: just construct a DVector
    // children who need more performance just overwrite
    return getScalarValue(new DVector(x, y));
  }
  
  /**
   * Gets the space dimension.
   */
  public int getSpaceDimension() {
    return dim;
  }
  
  /**
   * Sets the space dimension.
   */
  public void setSpaceDimension(int dim) {
    this.dim = dim;
  }
  
  /**
   * Gets position of last send request.
   */
  public DVector getLastPosition() {
    return lastPosition;
  }
  
 /**
   * Gets scalar value at position of last send request.
   */
  public double getLastScalarValue() {
    return getScalarValue(lastPosition);
  }
  
  /**
   * sends Vector event with scalar field value at a given position
   */
  public void sendScalarValue(DVector coord) {
    lastPosition = coord;
    fireVectorEvent(new VectorEvent(this, new DVector(getScalarValue(coord))));
  }
  
  // the usual support for Vector events
  protected transient VectorSupport vecSup;
  
  /** add a VectorListener */
  public synchronized void addVectorListener(VectorListener l) {
    vecSup.addVectorListener(l);
  }
  
  /** remove a VectorListener */
  public synchronized void removeVectorListener(VectorListener l) {
    vecSup.removeVectorListener(l);
  }
  
  /** fire a Vector event */
  public void fireVectorEvent(VectorEvent e) {
    vecSup.fireVectorEvent(e);
  }
}
