package physbeans.phys;

import physbeans.func.*;
import physbeans.math.*;
import java.io.*;

/**
 * A physical bean describing a mass point in two dimensions that is subject to
 * an arbitrary force F(x, y, vx, vy, t) according to the following ODE:<br>
 *    m x'' = Fx(x, y, vx, vy, t)
 *    m y'' = Fy(x, y, vx, vy, t)
 * <p>
 * It has the following parameters
 * <ul>
 * <li>mass m</li>
 * <li>force functions Fx, Fy defined by parsing two function strings,</li>
 * <li>a list of parameter names that can be used in the definition of Fx and
 *     Fy in addition to the standard variables x, y, vx, vy, t, and m,</li>
 * <li>values for the parameters.</li>
 * </ul>
 * <p>Generic2DMassPoint stores initial values and can be restarted.</p>
 * </p>
 */
public class Generic2DMassPoint extends SolveODEFunction implements Serializable {
  
  protected double m;
  protected String[] parameterNames;
  protected ParsedFunction parser;
  protected String[] varNames;
  protected DVector x0, v0;         // redundant, but needed to fix NetBeans bug
  
  /**
   * Creates a default instance of Generic1DMassPoint.
   * Describes a freely falling mass with g as single parameter.
   */
  public Generic2DMassPoint() {
    super();
    m = 1.0;
    varNames = new String[]{"x", "y", "vx", "vy", "t", "m"};
    
    // redundant, but needed to fix NetBeans bug
    // has to be consistent with NewtonODE !
    x0 = new DVector(1.0, 0.0);
    v0 = new DVector(0.0, 0.0);
    
    parser = new ParsedFunction();
    parser.setFunction(new String[]{"0", "-m*g"});
    parser.setParameter(new DVector(9.81));
    parser.setVariableName(new String[]{"x", "y", "vx", "vy", "t", "m", "g"});
    
    setOde(new NewtonODE());
    restart();
  }
  
  /**
   * Gets mass m attached to the spring.
   */
  public double getMass() {
    return this.m;
  }
  
  /**
   * Sets mass m attached to the spring.
   */
  public void setMass(double m) {
    this.m = m;
    restart();
  }
  
  /**
   * Gets description string of the x component of the force function.
   */
  public String getFunctionX() {
    return parser.getFunction(0);
  }
  
  /**
   * Sets description strings of the x component of the force function.
   */
  public void setFunctionX(String f) {
    parser.setFunction(0, f);
    restart();
  }
  
  /**
   * Gets description string of the y component of the force function.
   */
  public String getFunctionY() {
    return parser.getFunction(1);
  }
  
  /**
   * Sets description strings of the y component of the force function.
   */
  public void setFunctionY(String f) {
    parser.setFunction(1, f);
    restart();
  }
  
  /**
   * Gets the value of parameter i.
   */
  public double getParameter(int i) {
    return parser.getParameter(i);
  }
  
  /**
   * Gets the values of all parameters.
   */
  public DVector getParameter() {
    return parser.getParameter();
  }
  
  /**
   * Sets the value of parameter i.
   */
  public void setParameter(int i, double val) {
    parser.setParameter(i, val);
  }
  
  /**
   * Sets the values of all parameters.
   */
  public void setParameter(DVector vals) {
    parser.setParameter(vals);
  }
  
  /**
   * Gets the name of parameter i.
   */
  public String getParameterName(int i) {
    // the first variable names are fixed
    
    return parser.getVariableName(i + varNames.length);
  }
  
  /**
   * Gets the names of all parameters.
   */
  public String[] getParameterName() {
    // cut off the fixed variable names
    
    int vl = varNames.length;
    String[] vars = parser.getVariableName();
    String[] pars = new String[vars.length - vl];
    
    for (int i = 0; i < pars.length; i++) {
      pars[i] = vars[i + vl];
    }
    
    return pars;
  }
  
  /**
   * Sets the name of parameter i.
   */
  public void setParameterName(int i, String name) {
    parser.setVariableName(i + varNames.length, name);
  }
  
  /**
   * Sets the names of all parameters.
   */
  public void setParameterName(String[] names) {
    int vl = varNames.length;
    int nl = names.length;
    String[] vars = new String[vl + nl];
    
    for (int i = 0; i < vl; i++) {
      vars[i] = varNames[i];
    }
    
    for (int i = vl; i < vl + nl; i++) {
      vars[i] = names[i - vl];
    }
    
    parser.setVariableName(vars);
  }
  
  /**
   * Gets initial position x0 (a 2d vector).
   */
  public DVector getX0() {
    return x0;
    // return getInitialValues().getDVector(0, 1);
  }
  
  /**
   * Sets initial position x0 (a 2d vector).
   */
  public void setX0(DVector x0) {
    this.x0 = x0;       // redundant, but needed to fix NetBeans bug
    setInitialValue(0, x0.x());
    setInitialValue(1, x0.y());
  }
  
  /**
   * Gets initial velocity v0 (a 2d vector).
   */
  public DVector getV0() {
    return v0;
    // return getInitialValues().getDVector(2, 3);
  }
  
  /**
   * Sets initial velocity v0 (a 2d vector).
   */
  public void setV0(DVector v0) {
    this.v0 = v0;       // redundant, but needed to fix NetBeans bug
    setInitialValue(2, v0.x());
    setInitialValue(3, v0.y());
  }
  
  /**
   * Gets current value of the position vector.
   * <P>Conveniance function to simplify routing.
   */
  public DVector getPosition() {
    DVector xv = getOutputVector();
    return xv.getDVector(0, 1);
  }
  
  /**
   * Gets current values of the velocity vector.
   * <P>Conveniance function to simplify routing.
   */
  public DVector getVelocity() {
    DVector xv = getOutputVector();
    return xv.getDVector(2, 3);
  }
  
  class NewtonODE extends ODE {
    
    public NewtonODE() {
      super(new DVector(1.0, 0.0, 0.0, 0.0), 0.0);
    }
    
    public DVector f(DVector x, double t) {
      DVector arg = new DVector(6);
      arg.setDVector(0, 3, x);     // x, y, vx, vy
      arg.set(4, t);
      arg.set(5, m);
      parser.setInputVector(arg);
      DVector force = parser.getOutputVector();
      
      DVector result = new DVector(4);
      result.set(0, x.get(2));        // x' = vx
      result.set(1, x.get(3));        // y' = vy
      result.set(2, force.x() / m);     // vx' = Fx/m
      result.set(3, force.y() / m);     // vy' = Fy/m
      
      return result;
    }
  }
}
