package physbeans.phys;

import physbeans.math.*;
import static java.lang.Math.*;
import static physbeans.math.SpecialPolynomials.*;

/**
 * Creates an arbitrary superposition of eigenfunctions of the 2d harmonic
 * oscillator as initial values on a 2d complex grid.<br>
 * The superposition is defined by a complex matrix c<sub>i,j</sub>, which
 * contains the coefficients of the eigenvectors
 *     &phi;<sub>i</sub>(x) * &phi;<sub>j</sub>(y).
 */
public class Harmonic2DEigenfunction extends ComplexGridInitializer {

  protected CMatrix c;      // unnormalized coefficients of the eigen functions
  protected double om;
  //protected int n,  m;
  protected IVector nxy;
  protected double[] N;
  protected int n0, m0;

  public Harmonic2DEigenfunction() {
    //n = 2;
    //m = 2;
    nxy = new IVector(2, 2);
    c = new CMatrix(new double[]{0.5, 0.5, 0.5, 0.5}, new double[]{0, 0, 0, 0}, 2);
    om = 1.0e14;
    N = new double[]{};                   // normalisation constants
    n0 = 0;
    m0 = 0;
  }

  /**
   * Sets initial values for the ScalarField f.
   */
  public void setGrid() {
    if (field == null) {
      return;
    }

    int max = nxy.max();
    if (N.length < max) {                  // lazy initialisation
      computeNormalisationConstants(max);
    }

    double[][] gridR = field.getGridReal();
    double[][] gridI = field.getGridImag();

    int nx = field.getGridSizeX();
    int ny = field.getGridSizeY();
    int bwX = field.getBoundaryWidthX();
    int bwY = field.getBoundaryWidthY();

    clearBoundaries();

    double fac = sqrt(om * Constants.ME / Constants.HBAR);
    for (int i = bwX; i < nx - bwX; i++) {
      for (int j = bwY; j < ny - bwY; j++) {
        DVector xVec = field.getWorldFromIndex(i, j);
        gridR[j][i] = 0.0;
        gridI[j][i] = 0.0;

        for (int k = 0; k < nxy.get(0); k++) {
          for (int l = 0; l < nxy.get(1); l++) {
            double xarg = fac * xVec.x();
            double psix = N[k] * H(k, xarg) * exp(-0.5 * xarg * xarg);
            double yarg = fac * xVec.y();
            double psiy = N[l] * H(l, yarg) * exp(-0.5 * yarg * yarg);

            Complex c_kl = c.get(k, l);
            gridR[j][i] += c_kl.real() * psix * psiy;
            gridI[j][i] += c_kl.imag() * psix * psiy;
          }
        }
      }
    }

    normalizeField();
  }

  /**
   * Sets complex matrix of coefficients c_ij.
   */
  public void setMatrix(CMatrix c) {
    this.c = c;
    nxy.set(1, c.getColumnDimension());
    nxy.set(0, c.getRowDimension());

    updateField();
  }

  /**
   * Gets complex matrix of coefficients c_ij.
   */
  public CMatrix getMatrix() {
    return c;
  }

  /**
   * Sets the (circular) frequency of the harmonic potential.
   */
  public void setOmega(double om) {
    this.om = om;
    updateField();
  }

  /**
   * Gets the (circular) frequency of the harmonic potential.
   */
  public double getOmega() {
    return om;
  }

  /**
   * Sets number of eigenfunctions in x and y direction.
   * All coefficients are set to 1.0 / sqrt(nx * ny);
   */
  public void setNEigenfunctions(IVector nxy) {
    this.nxy = nxy;
    int n = nxy.get(0);
    int m = nxy.get(1);

    double val = 1.0 / sqrt(n * m);
    c = new CMatrix(new Matrix(n, m, val), new Matrix(n, m, 0.0));

    updateField();
  }

  /**
   * Gets number of eigenfunctions in x and y direction.
   */
  public IVector getNEigenfunctions() {
    return nxy;
  }

  /**
   * Sets the absolute values of the coefficients in x direction.
   * In y direction only the first eigenfunction is used.
   * 
   * This is a conveniance function for the special case
   *    dim(c) = N x 1
   *  i.e. in y direction only the ground state is used.
   */
  public void setAbsXCoefficients(DVector v) {
    if (nxy.get(0) != v.getDimension() || nxy.get(1) != 1) {
      setNEigenfunctions(new IVector(v.getDimension(), 1));
    }

    CVector colVec = c.getColumnVector(0);
    colVec.setAbs(v);
    c.setColumnVector(0, colVec);

    updateField();
  }

  /**
   * Sets the phase angles of the coefficients in x direction.
   * In y direction only the first eigenfunction is used.
   *
   * This is a conveniance function for the special case
   *    dim(c) = N x 1
   *  i.e. in y direction only the ground state is used.
   */
  public void setPhaseXCoefficients(DVector v) {
    if (nxy.get(0) != v.getDimension() || nxy.get(1) != 1) {
      setNEigenfunctions(new IVector(v.getDimension(), 1));
    }

    CVector colVec = c.getColumnVector(0);
    colVec.setPhase(v);
    c.setColumnVector(0, colVec);

    updateField();
  }

  /**
   * Fixes the x eigenvalue to n, but retains the y mixture.
   * This correspondends to a partial measurement of the x coefficient.
   */
  public void setXFunction(int k) {
    if (k < 0 || k > nxy.get(0)) {
      return;
    }
    CVector row = c.getRowVector(k);
    row.timesEquals(1.0 / row.norm());

    c = new CMatrix(nxy.get(0), nxy.get(1));   // empty matrix
    c.setRowVector(k, row);

    updateField();
  }

  /**
   * Fixes the y eigenvalue to n, but retains the x mixture.
   * This correspondends to a partial measurement of the y coefficient.
   */
  public void setYFunction(int k) {
    if (k < 0 || k > nxy.get(1)) {
      return;
    }
    CVector col = c.getColumnVector(k);
    col.timesEquals(1.0 / col.norm());

    c = new CMatrix(nxy.get(0), nxy.get(1));   // empty matrix
    c.setColumnVector(k, col);

    updateField();
  }

  /**
   * Sets coefficients to a simple eigenfunction given by new n0 and old m0.
   *
   * This is a conveniance function for the special case
   *   c<sub>i,j</sub> = 0 for i &ne; n0, j &ne; m0
   * i.e. only a single eigenfunction is used.
   */
  public void setSimpleEigenfunctionX(int n0) {
    this.n0 = n0;
    computeSimpleMatrix();
    updateField();
  }

  /**
   * Gets eigennumber used by in x direction.
   */
  public int getSimpleEigenfunctionX() {
    return n0;
  }

  /**
   * Sets coefficients to a simple eigenfunction given by old n0 and new m0.
   *
   * This is a conveniance function for the special case
   *   c<sub>i,j</sub> = 0 for i &ne; n0, j &ne; m0
   * i.e. only a single eigenfunction is used.
   */
  public void setSimpleEigenfunctionY(int m0) {
    this.m0 = m0;
    computeSimpleMatrix();
    updateField();
  }

  /**
   * Gets eigennumber used by in y direction.
   */
  public int getSimpleEigenfunctionY() {
    return m0;
  }

  /**
   * Computes the normalisation constants of the eigenfunctions up to n = max.
   */
  protected void computeNormalisationConstants(int max) {
    double fac = sqrt(om * Constants.ME / Constants.HBAR);
    N = new double[max];
    N[0] = sqrt(fac / sqrt(PI));

    for (int i = 1; i < max; i++) {
      N[i] = N[i - 1] / sqrt(2 * i);
    }
  }

  /**
   * Computes the coefficient matrix for a single eigenfunction
   * given by eigenvalues n0, m0.
   */
  protected void computeSimpleMatrix() {
    int n = max(n0 + 1, nxy.get(0));
    int m = max(m0 + 1, nxy.get(1));
    nxy = new IVector(n, m);

    c = new CMatrix(n, m);
    c.set(n0, m0, new Complex(1.0));
  }
}
