package physbeans.phys;

import java.util.*;

import physbeans.math.*;
import physbeans.model.*;

/**
 * Scalar field describing the temperature distribution of
 * three layers of thermally insulating materials
 * with different layer widths. <br>
 * Computes thermal flow through the layers <br>
 * sends events with new value for thermal flow<br>
 */
public class LayeredTemperatureField extends Grid2DScalarField {
  protected double d1, d2, d3;    // layer widths in m
  protected double lambda1, lambda2, lambda3; // layer thermal conductivities in W/(K m)
  protected double t0, t3;        // temperatures at the left and right edges, in K
  
  // derived values
  protected double t1, t2;  // temperatures at intermediate points, in K
  protected double phi;   // thermal flow per area, in W/m^2, phi < 0
  protected int i1, i2, i3, i4;   // grid indices of wall positions
  
  /**
   * LayeredTemperatureField with default values
   */
  public LayeredTemperatureField() {
    super();
    
    // grid related values
    nx = 400;
    ny = 2;
    xMin = -0.2;
    xMax = 0.65;
    yMin = -1.0;
    yMax = 1.0;
    bwX = 0;
    bwY = 0;
    
    // layer widths
    d1 = 0.1;
    d2 = 0.05;
    d3 = 0.1;
    
    // layer conductivities
    lambda1 = 0.6;
    lambda2 = 0.026;
    lambda3 = 0.6;
    
    // temperatures at the edges
    t0 = 293.0;
    t3 = 273;
    
    initGrid();
  }
  
  /**
   * computes the derived values t1, t2 and phi
   */
  protected void setupGrid() {
    super.setupGrid();
    phi = (t3 - t0)/(d1/lambda1 + d2/lambda2 + d3/lambda3);
    t1 = t0 + phi*d1/lambda1;
    t2 = t1 + phi*d2/lambda2;
    
    i1 = getIndexForX(0.0);
    i2 = getIndexForX(d1);
    i3 = getIndexForX(d1 + d2);
    i4 = getIndexForX(d1 + d2 + d3);
  }
  
  
  /**
   * get thermal flow per area
   */
  public double getPhi() {
    return -phi;  // phi is negative in my system
  }
  
  /**
   * get width of first layer
   */
  public double getD1() {
    return d1;
  }
  
  /**
   * set width of first layer
   */
  public void setD1(double d) {
    d1 = d;
    initGrid();
    trigger();
  }
  
  /**
   * get width of second layer
   */
  public double getD2() {
    return d2;
  }
  
  /**
   * set width of second layer
   */
  public void setD2(double d) {
    d2 = d;
    initGrid();
    trigger();  }
  
  /**
   * get width of third layer
   */
  public double getD3() {
    return d3;
  }
  
  /**
   * set width of third layer
   */
  public void setD3(double d) {
    d3 = d;
    initGrid();
    trigger();  }
  
  /**
   * get thermal conductivity of first layer
   */
  public double getLambda1() {
    return lambda1;
  }
  
  /**
   * set thermal conductivity of first layer
   */
  public void setLambda1(double l) {
    lambda1 = l;
    initGrid();
    trigger();  }
  
  /**
   * get thermal conductivity of second layer
   */
  public double getLambda2() {
    return lambda2;
  }
  
  /**
   * set thermal conductivity of second layer
   */
  public void setLambda2(double l) {
    lambda2 = l;
    initGrid();
    trigger();  }
  
  /**
   * get thermal conductivity of third layer
   */
  public double getLambda3() {
    return lambda3;
  }
  
  /**
   * set thermal conductivity of third layer
   */
  public void setLambda3(double l) {
    lambda3 = l;
    initGrid();
    trigger();  }
  
  /**
   * Gets temperature at the left edge, in K.
   */
  public double getT0() {
    return this.t0;
  }
  
  /**
   * Sets temperature at the left edge, in K.
   */
  public void setT0(double t0) {
    this.t0 = t0;
  }
  
  /**
   * Gets temperature at the right edge, in K.
   */
  public double getT3() {
    return this.t3;
  }
  
  /**
   * Sets temperature at the right edge, in K.
   */
  public void setT3(double t3) {
    this.t3 = t3;
  }
  
  /**
   * temperature varying in x through the layers
   */
  protected void computeGrid() {
    double t;
    for (int i = 0; i < i1; i++) {    // left side
      val[0][i] = t0;
      val[1][i] = t0;
    }
    
    for (int i = i1; i < i2; i++) {    // inside first layer
      t = t0 + (i - i1)*(t1 - t0)/(i2 - i1) ;
      val[0][i] = t;
      val[1][i] = t;
    }
    
    for (int i = i2; i < i3; i++) {    // inside second layer
      t = t1 + (i - i2)*(t2 - t1)/(i3 - i2);
      val[0][i] = t;
      val[1][i] = t;
    }
    
    for (int i = i3; i < i4; i++) {    // inside third layer
      t = t2 + (i - i3)*(t3 - t2)/(i4 - i3);
      val[0][i] = t;
      val[1][i] = t;
    }
    
    for (int i = i4; i < nx; i++) {    // right side
      val[0][i] = t3;
      val[1][i] = t3;
    }
  }
  

}
