package physbeans.phys;

import physbeans.math.*;

/**
 * Abstract base class of all light sources. Has a (x) position at the optical
 * bench and a height (y) above the optical axis and produces a LightRayBundle.
 */
abstract public class LightSource extends PhysModel implements Comparable {
  
  protected double pos;
  protected double height;
  protected LightRayBundle rays;
  
  /** Creates a new instance of LightSource */
  public LightSource() {
    super();
    pos = 0.0;
    height = 0.0;
    rays = new LightRayBundle();
  }
  
  /**
   * Gets the position (the x coordinate of a specified point) of the
   * light source.
   */
  public double getPosition() {
    return pos;
  }
  
  /**
   * Sets the position (the x coordinate of a specified point) of the
   * light source.
   */
  public void setPosition(double pos) {
    this.pos = pos;
    trigger();
  }
  
  /**
   * Gets the height of the light source over the optical axis.
   */
  public double getHeight() {
    return this.height;
  }
  
  /**
   * Sets the height of the light source over the optical axis.
   */
  public void setHeight(double height) {
    this.height = height;
    trigger();
  }
  
  /**
   * Gets the position and height of the light source as a vector (pos, height).
   */
  public DVector getCoords() {
    return new DVector(pos, height);
  }
  
  /**
   * Sets the position and height of the light source from a vector (pos, height).
   */
  public void setCoords(DVector p) {
    pos = p.get(0);
    height = p.get(1);
    trigger();
  }
  
  /**
   * Gets the current light rays of the light source.
   */
  public LightRayBundle getRays() {
    return rays;
  }
  
  /**
   * Gets new light rays from the light source.
   */
  public LightRayBundle getNewRays() {
    computeRays();
    return rays;
  }
  
  /**
   * Recompute rays and inform listeners.
   */
  public void trigger() {
    computeRays();
    super.trigger();
  }
  
  /**
   * Uses the position along the optical (x) axis to compare LightSources.
   * Returns 1 (0, -1), if this object has a larger (equal, smaller) x position
   * than the specified object o.
   */
  public int compareTo(Object o) {
    LightSource el = (LightSource) o;
    if (Maths.isStrictlyPositive(pos - el.pos)) {
      return 1;
    } else if (Maths.isStrictlyPositive(el.pos - pos)) {
      return -1;
    } else {
      return 0;
    }
  }
  
  /**
   * Computes the light ray bundle emanating from the source.
   */
  abstract protected void computeRays();
}
