package physbeans.phys;

import physbeans.math.*;

/**
 * An ObjectLightSource creates a bundle of rays that are used for the
 * construction of an image. For this purpose it has a reference to an
 * ImagingObject, which defines a focal point. The bundle consists of four rays:
 * a parallel ray, a focal ray, a central ray and an additional ray with an
 * arbitrary direction.
 */
public class ObjectLightSource extends LightSource {
  
  ImagingElement refElement;
  protected double angle;
  
  /** Creates a new instance of ObjectLightSource */
  public ObjectLightSource() {
    super();
    angle = 0.0;
    trigger();
  }
  
  protected void computeRays() {
    if (refElement == null) {
      return;
    }
    
    rays = new LightRayBundle();
    DVector angles = getStartAngles();
    
    for (int i = 0; i < 4; i++) {
      rays.addRay(new LightRay(pos, height, angles.get(i)));
    }
  }
  
  /**
   * Gets the angle of the arbitrary ray against the optical axis.
   */
  public double getAngle() {
    return this.angle;
  }
  
  /**
   * Sets the angle of the arbitrary ray against the optical axis.
   */
  public void setAngle(double angle) {
    this.angle = angle;
    trigger();
  }
  
  /**
   * Gets the reference element for the object light source. It defines the
   * position of the focal point.
   */
  public ImagingElement getRefElement() {
    return this.refElement;
  }
  
  /**
   * Sets the reference element for the object light source. It defines the
   * position of the focal point.
   */
  public void setRefElement(ImagingElement refElement) {
    this.refElement = refElement;
    trigger();
  }
  
  /**
   * Returns the start angles of the four rays starting at the object source,
   * in the order parallel, focal, central, arbitrary.
   */
  public DVector getStartAngles() {
    if (refElement == null) {
      return null;
    }
    
    double f = refElement.getFocalPoints().getPoint(0).get(0);
    double c = refElement.getPosition();
    
    double alp = (pos < c)? 0.0 : Math.PI;
    double alf = Math.atan2(-height, f - pos);
    if ((c -pos)*(pos - f) > 0) {     // object is between mirror and focal point
      alf += Math.PI;
      if (alf > Math.PI) {
        alf -= 2*Math.PI;      // reduce to fundamental interval
      }
    }
    double alc = Math.atan2(-height, c - pos);
    
    return new DVector(alp, alf, alc, angle);
  }
}
