package physbeans.phys;

import physbeans.math.*;

/**
 * Class containing physical laws of reflection and refraction.
 */
public class OpticalSurface {
  
  /**
   * Returns the reflection angle for given incoming and normal vector.
   * Both vectors are assumed to be of length 1.
   */
  static public double getReflectionAngle(DVector in, DVector n) {
    DVector par = n.times(in.times(n));
    DVector out = in.minus(par.times(2));
    return Math.atan2(out.y(), out.x());
  }
  
  /**
   * Returns the refraction angle for given incoming and normal vector. 
   * The index of refraction at the incoming (outgoing) side is ni (no).
   * Both vectors are assumed to be of length 1, the normal vector points 
   * to the side with nIn.
   */
  static public double getRefractionAngle(DVector inVec, DVector nVec,
      double ni, double no) {
    // formulae from Wiedemann, Numerische Physik, p.102f.  are wrong!
    // beware: direction of nVec is outwards, in Wiedemann inwards!
    
    double xi = inVec.times(nVec);
    double l1;
    if (Maths.isStrictlyPositive(xi)) {   // ray is leaving
      l1 = ni/no;
    } else {                              // ray is entering
      l1 = ni/no;
      nVec.timesEquals(-1.0);
      xi = -xi;
    }
    
    double l2 = Math.sqrt(1 - l1*l1*(1 - xi*xi)) - xi*l1;
    DVector out;
    if (Double.isNaN(l2)) {
      DVector par = nVec.times(inVec.times(nVec));
      out = inVec.minus(par.times(2));
    } else {
      out = inVec.times(l1);
      out.plusEquals(nVec.times(l2));
    }
    
    return Math.atan2(out.y(), out.x());
  }
}
