package physbeans.phys;

import physbeans.math.*;
import java.awt.*;

/**
 * A PointWaveSource disturbs a given point in the ripple tank by setting it
 * periodically according to a harmonic function.<br>
 * Additionally one can define a (two-dimensional) velocity, which defines the
 * motion of the PointWaveSource.
 */
public class PointWaveSource extends WaveSource {
  
  protected  DVector pos0, v;
  protected double a, om, phi;
  
  protected DVector pos;
  protected double tLast;
  
  /** Creates a new instance of PointWaveSource */
  public PointWaveSource() {
    super();
    pos0 = new DVector(0.0, 0.0);
    v = new DVector(0.0, 0.0);
    a = 1.0;
    om = 1.0;
    phi = 0.0;
    
    pos = new DVector(0.0, 0.0);
    tLast = 0.0;
  }
  
  public void stirTank(RippleTank tank, double t) {
    pos = pos.plus(v.times(t - tLast));
    tLast = t;
    Point p = tank.getIndex(pos);
    if ((p.x < 0) || (p.x >= tank.getGridSizeX()) ||
        (p.y < 0) || (p.y >= tank.getGridSizeY())) {
      return;
    }
    
    double[][] grid = tank.getGrid();
    grid[p.y][p.x] = a*Math.cos(om*t + phi);
  }
  
  /**
   * Gets the initial position of the source.
   */
  public DVector getInitialPosition() {
    return pos0;
  }
  
  /**
   * Sets the initial position of the source.<br>
   * Moves the source to the position as well.
   */
  public void setInitialPosition(DVector p) {
    pos0 = p;
    pos = pos0.copy();
  }
  
  
  /**
   * Gets the velocity of the source.
   */
  public DVector getVelocity() {
    return v;
  }
  
  /**
   * Sets the velocity of the source.
   */
  public void setVelocity(DVector v) {
    this.v = v;
  }
  
  /**
   * Sets the x component of the source velocity.<br>
   * Conveniance method to simplify routing.
   */
  public void setVelocityX(double vx) {
    v.set(0, vx);
  }
  
  /**
   * Sets the y component of the source velocity.<br>
   * Conveniance method to simplify routing.
   */
  public void setVelocityY(double vy) {
    v.set(1, vy);
  }
  
  
  /**
   * Gets the amplitude of the source.
   */
  public double getAmplitude() {
    return a;
  }
  
  /**
   * Sets the amplitude of the source.
   */
  public void setAmplitude(double a1) {
    a = a1;
  }
  
  /**
   * Gets the frequency of the source.
   */
  public double getFrequency() {
    return om/(2*Math.PI);
  }
  
  /**
   * Sets the frequency of the source.
   */
  public void setFrequency(double f) {
    om = 2*Math.PI*f;
  }
  
  /**
   * Gets the phase of the source.
   */
  public double getPhase() {
    return phi;
  }
  
  /**
   * Sets the phase of the source.
   */
  public void setPhase(double p) {
    phi = p;
  }
  
  /**
   * Sets current source position to initial value.
   */
  public void restart() {
    pos = pos0.copy();
    tLast = 0.0;
  }
}
