package physbeans.phys;

import physbeans.math.*;
import static java.lang.Math.*;

/**
 * Creates a gaussian wave packet in polar coordinates as initial value on a
 * 2d complex grid.
 * Explicitely it is given by
 *   psi(r, phi) = N * exp(-0.5*((r-r0)/dr)^2)* exp(-0.5*((phi-phi0)/dphi)^2)
 *                    * exp(i/hbar * lz * phi)
 */
public class PolarGaussianGridInitializer extends ComplexGridInitializer {

  protected double r0, phi0, dr, dphi, lz;

  public PolarGaussianGridInitializer() {
    super();
    r0 = 1.0;
    phi0 = 0.0;
    dr = 0.1;
    dphi = 0.5;
    lz = 0.0;
  }

  /**
   * Sets initial values for the ScalarField f.
   */
  public void setGrid() {
    if (field == null) {
      return;
    }

    double[][] gridR = field.getGridReal();
    double[][] gridI = field.getGridImag();

    int nx = field.getGridSizeX();
    int ny = field.getGridSizeY();
    int bwX = field.getBoundaryWidthX();
    int bwY = field.getBoundaryWidthY();

    clearBoundaries();

    for (int i = bwX; i < nx - bwX; i++) {
      for (int j = bwY; j < ny - bwY; j++) {
        DVector x = field.getWorldFromIndex(i, j);

        double r = hypot(x.x(), x.y());
        double phi = atan2(x.y(), x.x());
        double rs = (r - r0) / dr;
        double phis = IEEEremainder((phi - phi0), 2 * PI) / dphi;

        double xr = exp(-0.5 * rs * rs) * exp(-0.5 * phis * phis);
        double xphi = (lz / Constants.HBAR) * phi;

        gridR[j][i] = xr * cos(xphi);
        gridI[j][i] = xr * sin(xphi);
      }
    }

    normalizeField();
  }

  /**
   * Sets initial distance r0 of the wave packet.
   */
  public void setR0(double r) {
    r0 = r;
    updateField();
  }

  /**
   * Gets initial distance r0 of the wave packet.
   */
  public double getR0() {
    return r0;
  }

  /**
   * Sets initial angle phi0 of the wave packet.
   */
  public void setPhi0(double phi) {
    phi0 = phi;
    updateField();
  }

  /**
   * Gets initial angle phi0 of the wave packet.
   */
  public double getPhi0() {
    return phi0;
  }

  /**
   * Sets radial width (standard deviation) of the wave packet.
   */
  public void setRadialWidth(double drNew) {
    dr = drNew;
    updateField();
  }

  /**
   * Gets radial width (standard deviation) of the wave packet.
   */
  public double getRadialWidth() {
    return dr;
  }

  /**
   * Sets angular width (standard deviation) of the wave packet.
   */
  public void setAngularWidth(double dphiNew) {
    dphi = dphiNew;
    updateField();
  }

  /**
   * Gets angular width (standard deviation) of the wave packet.
   */
  public double getAngularWidth() {
    return dphi;
  }

  /**
   * Sets angular momentum (its mean value) of the wave packet.
   */
  public void setLz(double lz) {
    this.lz = lz;
    updateField();
  }

  /**
   * Gets angular momentum (its mean value) of the wave packet.
   */
  public double getLz() {
    return lz;
  }
}
