package physbeans.phys;

import java.util.*;

/**
 * A RippleTank simulates the behaviour of ideal two-dimensional waves in a
 * rectangular domain. Absorbing boundary conditions are used to create the
 * illusion of a virtually unbounded region.
 * <p>A RippleTank can have a number of WaveSources, which disturb the surface
 * in a time dependent way.
 * <p>Additionally a set of grid walls and regions with a different wave
 * velocity (a different medium) can be inserted.
 * <p>Since the time increment is defined by the grid and the medium (by fixing
 * the Courant constant xi), the RippleTank defines its time internally, using
 * the external Timer only for stepping. If the Timer values are used
 * explicitely in another bean, the Timer increment has to be set manually to
 * the exact value defined by the RippleTank. For this purpose it is displayed
 * as a bean property.
 */
public class RippleTank extends Grid2DScalarField {
  
  protected Vector<WaveSource> sources;
  protected WaveWall wall;
  protected WaveMedium medium;
  protected double c;
  protected double tau;
  
  protected double xi;        // Courant's constant, xi = c dt /dx
  protected double[][] damp;
  protected double[][] val1, val2;   // values of the two last time steps
  double[][] gridW, gridM;    // grids used of current wall and medium
  protected double t;
  // doesn't use Timer time values, because dt has to be computed here
  
  /** Creates a new instance of RippleTank */
  public RippleTank() {
    super();
    sources = new Vector<WaveSource>();
    wall = null;
    medium = null;
    
    c = 1.0;
    tau = 0.0;
    xi = 0.5;    // largest possible value: 1/sqrt(2)
    
    nx = 140;
    ny = 140;
    bwX = 20;
    bwY = 20;
    square = true;
    
    setupGrid();
  }
  
  /**
   * Computes all auxiliary grids and variables for a change of the grid
   * structure.
   */
  protected void setupGrid() {
    super.setupGrid();
    if ((val1 == null) || (val1[0].length != nx) || (val1.length != ny)) {
      val1 = new double[ny][nx];
      val2 = new double[ny][nx];
      damp = new double[ny][nx];
    }
    
    createAbsorbingBoundary();
    
    // if no wall is given an empty wall is created.
    // this simplifies the inner loop and increases performance.
    if (wall == null) {
      wall = new EmptyWall();
    }
    wall.setup(this);
    gridW = wall.getGrid();
    
    if (medium != null) {
      medium.setup(this);
      gridM = medium.getGrid();
    }
    
    t = 0.0;
  }
  
  
  protected double left, right, up, down;
  
  /**
   * Computes the next grid values.
   */
  protected void computeGrid() {
    // time shift
    double[][] temp = val2;
    val2 = val1;
    val1 = val;
    val = temp;
    interp.setGrid(val);
    
    double xi2 = xi*xi;
    double v, al;
    
    // for performance reasons the medium check is performed outside the loop
    if (medium == null) {
      for (int i = 1; i < nx - 1; i++) {
        for (int j = 1; j < ny - 1; j++) {
          if (gridW[j][i] == 1.0) {
            continue;
          }
          
          getNeighbours(i, j);
          
          v = xi2*(up + down + left + right);
          al = 1.0/damp[j][i] - 1.0;
          v += (2 + al - 4*xi2)*val1[j][i] - val2[j][i];
          val[j][i] = damp[j][i]*v;
        }
      }
    } else {
      // Instead of a Yee like algorithm with c_i,j living on the edges I use a
      // very simple discretisation scheme. It should work for my "almost
      // homogeneous" cases. This corresponds to a wave equation with a term
      // c^2 d_x^2 u  instead of  d_x (c^2 d_x u).
      // Furthermore: I use the value of the time increment computed from the
      // global c, which therefore should be at least the maximal value of c_i,j.
      
      double xi2medium = xi2/(c*c);
      for (int i = 1; i < nx - 1; i++) {
        for (int j = 1; j < ny - 1; j++) {
          if (gridW[j][i] == 1.0) {
            continue;
          }
          
          getNeighbours(i, j);
          
          v = xi2medium*(gridM[j-1][i]*up + gridM[j+1][i]*down +
              gridM[j][i-1]*left + gridM[j][i+1]*right);
          al = 1.0/damp[j][i] - 1.0;
          v += (2 + al - 4*xi2medium*gridM[j][i])*val1[j][i] - val2[j][i];
          val[j][i] = damp[j][i]*v;
        }
      }
    }
    
    setSourceValues(t);
  }
  
  
  /**
   * Gets values for the neighbour fields during a grid iteration.
   * Respects surrounding walls.
   */
  protected void getNeighbours(int i, int j) {
    if (gridW[j][i-1] == 1.0) {
      left = wall.isFreeEnd() ? val1[j][i] : 0.0;
    } else {
      left = val1[j][i-1];
    }
    
    if (gridW[j][i+1] == 1.0) {
      right = wall.isFreeEnd() ? val1[j][i] : 0.0;
    } else {
      right = val1[j][i+1];
    }
    
    if (gridW[j-1][i] == 1.0) {
      up = wall.isFreeEnd() ? val1[j][i] : 0.0;
    } else {
      up = val1[j-1][i];
    }
    
    if (gridW[j+1][i] == 1.0) {
      down = wall.isFreeEnd() ? val1[j][i] : 0.0;
    } else {
      down = val1[j+1][i];
    }
  }
  
  /**
   * Proceeds the state of the grid by one time step.<br>
   * The step size is given by the value of xi, the Courant constant.
   */
  public void nextStep() {
    computeGrid();
    trigger();
    t += getTimeIncrement();
  }
  
  /**
   * Sets system to start values and prepares stepping.
   */
  public void restart() {
    setInitialConditions();
    t = 0.0;
    trigger();
  }
  
  /**
   * Implements the initial conditions on all grids.
   */
  protected void setInitialConditions() {
    // TODO: other possible initial conditions
    for (int i = 0; i < nx; i++) {
      for (int j = 0; j < ny; j++) {
        val[j][i] = 0.0;
        val1[j][i] = 0.0;
        val2[j][i] = 0.0;
      }
    }
  }
  
  /**
   * Sets the grid points with sources to their values at time t.
   */
  protected void setSourceValues(double t) {
    for (int nSource = 0; nSource < sources.size(); nSource++) {
      if (sources.get(nSource) != null) {
        ((WaveSource) sources.get(nSource)).stirTank(this, t);
      }
    }
  }
  
  /**
   * Sets up the grid of damping values to implement an absorbing layer.
   */
  protected void createAbsorbingBoundary() {
    // set interior
    double di = 1/(1 + tau*getTimeIncrement());
    for (int i = bwX; i < nx - bwX; i++) {
      for (int j = bwY; j < ny - bwY; j++) {
        damp[j][i] = di;
      }
    }
    
    double d0 = 0.999*di;
    double q = 0.01;
    // set boundary columns
    for (int i = 0; i < bwX; i++) {
      double d = d0 - (bwX-i)*q;
      for (int j = i; j < ny-i; j++) {
        damp[j][i] = d;
        damp[j][nx-i-1] = d;
      }
    }
    
    // set boundary rows
    for (int j = 0; j < bwY; j++) {
      double d = d0 - (bwY-j)*q;
      for (int i = j; i < nx-j; i++) {
        damp[j][i] = d;
        damp[ny-j-1][i] = d;
      }
    }
  }
  
  /**
   * Sets grid size in x direction in pixel.
   * It has to be large enough to hold the boundary layer.
   */
  public void setGridSizeX(int n) {
    if (n <= 2*bwX) {
      return;
    }
    super.setGridSizeX(n);
  }
  
  /**
   * Sets grid size in y direction in pixel.
   * It has to be large enough to hold the boundary layer.
   */
  public void setGridSizeY(int n) {
    if (n <= 2*bwY) {
      return;
    }
    super.setGridSizeY(n);
  }
  
  /**
   * Adds a source to the list of wave sources.
   */
  public void addSource(WaveSource v) {
    if (!sources.contains(v)) {
      sources.add(v);
    }
    trigger();
  }
  
  /**
   * Gets number of wave sources.
   */
  public int getNSources() {
    return sources.size();
  }
  
  /**
   * Gets the velocity of wave propagation.
   */
  public double getWaveVelocity() {
    return c;
  }
  
  /**
   * Sets the velocity of wave propagation.
   * <P>If a medium is given, this value is only used to compute the time
   * increment. Therefore it should be set to the maximal wave velocity in the
   * medium (at least).
   */
  public void setWaveVelocity(double c1) {
    c = c1;
    trigger();
  }
  
  /**
   * Gets current time value.
   */
  public double getTime() {
    return t;
  }
  
  /**
   * Gets damping value at inner grid points.<br>
   * The damping value d is directly related to the damping constant tau of the
   * telegraph equation via
   *   d = 1/(1 + tau*Delta_t)
   */
  public double getDamping() {
    return tau;
  }
  
  /**
   * Sets damping value at inner grid points.<br>
   * The damping value d is directly related to the damping constant tau of the
   * telegraph equation via
   *   d = 1/(1 + tau*Delta_t)
   */
  public void setDamping(double tau1) {
    tau= tau1;
    createAbsorbingBoundary();
  }
  
  
  /**
   * Gets the currently used wall.
   */
  public WaveWall getWall() {
    return wall;
  }
  
  /**
   * Insert a new wall.
   */
  public void setWall(WaveWall w) {
    if (w != null) {
      wall = w;
      wall.setup(this);
    } else {
      wall = new EmptyWall();
    }
    gridW = wall.getGrid();
    
    trigger();
  }
  
  /**
   * Gets the currently used medium.
   */
  public WaveMedium getMedium() {
    return medium;
  }
  
  /**
   * Insert a new medium.
   */
  public void setMedium(WaveMedium m) {
    if (m != null) {
      medium = m;
      medium.setup(this);
      gridM = medium.getGrid();
    }
    
    trigger();
  }
  
  /**
   * Gets the value of the time increment used for the wave propagation.<br>
   * The time increment is computed as
   *      dt = xi*dx/c
   * where the Courant constant xi has the default value 0.5.
   */
  public double getTimeIncrement() {
    return xi*Math.min(getDX(), getDY())/c;
  }
  
}

// utility class
class EmptyWall extends WaveWall {
  protected void setGrid() {
  }
  
}
