package physbeans.phys;

import static java.lang.Math.*;

/**
 * Schroedinger2D simulates the two-dimensional Schrödinger equation on a
 * rectangular domain.
 */
public class Schroedinger2D extends ComplexGrid2DScalarField {

  protected ComplexGridInitializer initFunction;
  protected Schroedinger2DPotential V;
  protected double t;
  // doesn't use Timer time values, because dt has to be computed here
  protected double hbar;
  protected double me;
  protected double dt0;

  // cached values
  protected double p;
  protected double tStep[];
  protected double gridV[][];
  protected double cosV[][][],  sinV[][][];

  /** Creates a new instance of Schroedinger2D */
  public Schroedinger2D() {
    super();

    hbar = Constants.HBAR;
    me = Constants.ME;

    nx = 140;
    ny = 140;
    bwX = 20;
    bwY = 20;
    square = true;
    V = null;

    dt0 = 0.0;      // value is computed from grid data

    setupGrid();
  }

  /**
   * Computes all auxiliary values for a change of the grid
   * structure.
   */
  protected void setupGrid() {
    super.setupGrid();
    computeCachedMatrices();
    setInitalValues();
    t = 0.0;
  }

  /**
   * Computes the grid values after one time step.
   * Uses the fourth-order method by De Raedt and Michielsen.
   */
  protected void computeGrid() {
    applyU4();
  }

  /**
   * Proceeds the state of the grid by one time step.<br>
   * The step size is given by the value of xi, the Courant constant.
   */
  public void nextStep() {
    computeGrid();
    trigger();
    t += getTimeIncrement();
  }

  /**
   * Sets system to start values and prepares stepping.
   */
  public void restart() {
    setInitalValues();
    t = 0.0;
    trigger();
  }

  /**
   * Gets current time value.
   */
  public double getTime() {
    return t;
  }

  /**
   * Sets the value of the time increment.
   * If it is set to 0.0, it is computed from the grid data.
   */
  public void setTimeIncrement(double dtNew) {
    dt0 = dtNew;
    setupGrid();
  }

  /**
   * Gets the value of the time increment.
   */
  public double getTimeIncrement() {
    if (dt0 == 0.0) {
      return 1.5 * me / hbar * getDX() * getDY();
    } else {
      return dt0;
    }
  }

  /**
   * Gets function that defines the initial values.
   */
  public ComplexGridInitializer getInitFunction() {
    return initFunction;
  }

  /**
   * Sets function that defines the initial values.
   */
  public void setInitFunction(ComplexGridInitializer iFunc) {
    if (iFunc != null) {
      initFunction = iFunc;
      initFunction.setField(this);
      trigger();
    }
  }

  /**
   * Sets potential function.
   * Default (null) defines a free particle.
   */
  public void setPotential(Schroedinger2DPotential V) {
    if (V != null) {
      this.V = V;
      V.setup(this);
      gridV = V.getGrid();
    }
    computeCachedMatrices();
    trigger();
  }

  /**
   * Gets potential function.
   * Default (null) defines a free particle.
   */
  public Schroedinger2DPotential getPotential() {
    return V;
  }

  /**
   * Gets grid with values of the potential,
   */
  public double[][] getPotentialGrid() {
    return gridV;
  }

  /**
   * Implements the initial conditions.
   */
  protected void setInitalValues() {
    if (initFunction != null) {
      initFunction.setGrid();
    } else {
      // clear everything
      for (int i = 0; i < nx; i++) {
        for (int j = 0; j < ny; j++) {
          val[j][i] = 0.0;
          valI[j][i] = 0.0;
        }
      }
    }
  }

  /**
   * Computes  chi = U4(dt) psi
   * for the fourth-order approximant U4 of exp(-i*H*dt/HBAR).
   */
  protected void applyU4() {
    // time step sizes are precomputed and cached.

    applyU2(0);
    applyU2(0);
    applyU2(1);
    applyU2(0);
    applyU2(0);
  }

  /**
   * Computes  psi = U2(dt) psi
   * for the second-order approximant U2 of exp(-i*H*dt/HBAR).
   */
  protected void applyU2(int tIndex) {
    applyU1(tIndex);
    applyU1T(tIndex);
  }

  /**
   * Computes  psi = U1(dt) psi
   * for the first-order approximant U1 of exp(-i*H*dt/HBAR).
   */
  protected void applyU1(int tIndex) {
    double dx = getDX();
    double dy = getDY();
    double dt = tStep[tIndex];

    // exp(-i*H1*dt/HBAR)
    double al = -hbar / (24 * me * dx * dx);
    double ca = cos(al * dt);
    double sa = sin(al * dt);
    for (int ix = 2; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }
    for (int ix = 3; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }

    // exp(-i*H2*dt/HBAR)
    al = -hbar / (24 * me * dx * dx);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int ix = 4; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }
    for (int ix = 5; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }

    // exp(-i*H3*dt/HBAR)
    al = 2 * hbar / (3 * me * dx * dx);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int ix = 2; ix < nx - 3; ix += 2) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 1, 0);
      }
    }

    // exp(-i*H4*dt/HBAR)
    al = 2 * hbar / (3 * me * dx * dx);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int ix = 3; ix < nx - 3; ix += 2) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 1, 0);
      }
    }

    // exp(-i*H5*dt/HBAR)
    al = -hbar / (24 * me * dy * dy);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int iy = 2; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }
    for (int iy = 3; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }

    // exp(-i*H6*dt/HBAR)
    al = -hbar / (24 * me * dy * dy);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int iy = 4; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }
    for (int iy = 5; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }

    // exp(-i*H7*dt/HBAR)
    al = 2 * hbar / (3 * me * dy * dy);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int iy = 2; iy < ny - 3; iy += 2) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 1);
      }
    }

    // exp(-i*H8*dt/HBAR)
    al = 2 * hbar / (3 * me * dy * dy);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int iy = 3; iy < ny - 3; iy += 2) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 1);
      }
    }

    // exp(-i*H9*dt/HBAR)
    // use cached values
    for (int ix = 2; ix < nx - 2; ix++) {
      for (int iy = 2; iy < ny - 2; iy++) {
        double chiRe = cosV[tIndex][iy][ix] * val[iy][ix] -
                sinV[tIndex][iy][ix] * valI[iy][ix];
        double chiIm = cosV[tIndex][iy][ix] * valI[iy][ix] +
                sinV[tIndex][iy][ix] * val[iy][ix];
        val[iy][ix] = chiRe;
        valI[iy][ix] = chiIm;
      }
    }
  }

  /**
   * Computes  psi = U1^T(dt) psi
   * for the transpose of the first-order approximant U1 of exp(-i*H*dt/HBAR).
   */
  protected void applyU1T(int tIndex) {
    double dx = getDX();
    double dy = getDY();
    double dt = tStep[tIndex];

    // exp(-i*H9*dt/HBAR)
    for (int ix = 2; ix < nx - 2; ix++) {
      for (int iy = 2; iy < ny - 2; iy++) {
        double chiRe = cosV[tIndex][iy][ix] * val[iy][ix] -
                sinV[tIndex][iy][ix] * valI[iy][ix];
        double chiIm = cosV[tIndex][iy][ix] * valI[iy][ix] +
                sinV[tIndex][iy][ix] * val[iy][ix];
        val[iy][ix] = chiRe;
        valI[iy][ix] = chiIm;
      }
    }

    // exp(-i*H8*dt/HBAR)
    double al = 2 * hbar / (3 * me * dy * dy);
    double ca = cos(al * dt);
    double sa = sin(al * dt);
    for (int iy = 3; iy < ny - 3; iy += 2) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 1);
      }
    }

    // exp(-i*H7*dt/HBAR)
    al = 2 * hbar / (3 * me * dy * dy);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int iy = 2; iy < ny - 3; iy += 2) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 1);
      }
    }

    // exp(-i*H6*dt/HBAR)
    al = -hbar / (24 * me * dy * dy);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int iy = 4; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }
    for (int iy = 5; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }

    // exp(-i*H5*dt/HBAR)
    al = -hbar / (24 * me * dy * dy);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int iy = 2; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }
    for (int iy = 3; iy < ny - 4; iy += 4) {
      for (int ix = 2; ix < nx - 2; ix++) {
        applyUalpha(ca, sa, ix, iy, 0, 2);
      }
    }

    // exp(-i*H4*dt/HBAR)
    al = 2 * hbar / (3 * me * dx * dx);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int ix = 3; ix < nx - 3; ix += 2) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 1, 0);
      }
    }

    // exp(-i*H3*dt/HBAR)
    al = 2 * hbar / (3 * me * dx * dx);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int ix = 2; ix < nx - 3; ix += 2) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 1, 0);
      }
    }

    // exp(-i*H2*dt/HBAR)
    al = -hbar / (24 * me * dx * dx);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int ix = 4; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }
    for (int ix = 5; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }

    // exp(-i*H1*dt/HBAR)
    al = -hbar / (24 * me * dx * dx);
    ca = cos(al * dt);
    sa = sin(al * dt);
    for (int ix = 2; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }
    for (int ix = 3; ix < nx - 4; ix += 4) {
      for (int iy = 2; iy < ny - 2; iy++) {
        applyUalpha(ca, sa, ix, iy, 2, 0);
      }
    }
  }

  /**
   * Computes psi = Ualpha(dt) psi
   * where Ualpha is one of the (many) 2x2 matrices which make up U1.
   * Instead of the parameters al and dt the values
   *    ca = cos(al*dt) and sin(al*dt)
   * are used as arguments. They are computed in U1 so that they can be hoisted
   * above the index loops.
   */
  protected void applyUalpha(double ca, double sa, int iX, int iY, int sx, int sy) {
    double chi1Re = ca * val[iY][iX] - sa * valI[iY + sy][iX + sx];
    double chi1Im = ca * valI[iY][iX] + sa * val[iY + sy][iX + sx];
    double chi2Re = -sa * valI[iY][iX] + ca * val[iY + sy][iX + sx];
    double chi2Im = sa * val[iY][iX] + ca * valI[iY + sy][iX + sx];

    val[iY][iX] = chi1Re;
    valI[iY][iX] = chi1Im;
    val[iY + sy][iX + sx] = chi2Re;
    valI[iY + sy][iX + sx] = chi2Im;
  }

  /**
   * Pre-computes matrices to speed up the time step.
   */
  protected void computeCachedMatrices() {
    // one needs matrices for 2 different time steps (in U2!)

    p = 1.0 / (4.0 - pow(4.0, 1.0 / 3.0));  // time step parameter in U2

    double dx = getDX();
    double dy = getDY();
    double al = -5 * hbar * hbar / (me * (dx * dx + dy * dy));
    double dt = getTimeIncrement();

    tStep = new double[]{0.5 * p * dt, 0.5 * (1 - 4 * p) * dt};
    gridV = new double[ny][nx];
    cosV = new double[2][ny][nx];
    sinV = new double[2][ny][nx];

    if (V == null) {
      for (int tIndex = 0; tIndex < tStep.length; tIndex++) {
        double cVal = cos((al) * tStep[tIndex] / hbar);
        double sVal = sin((al) * tStep[tIndex] / hbar);
        for (int ix = bwX; ix < nx - bwX; ix++) {
          for (int iy = bwY; iy < ny - bwY; iy++) {
            cosV[tIndex][iy][ix] = cVal;
            sinV[tIndex][iy][ix] = sVal;
          }
        }
      }
    } else {
      gridV = V.getGrid();
      if (gridV == null) return;

      for (int ix = bwX; ix < nx - bwX; ix++) {
        for (int iy = bwY; iy < ny - bwY; iy++) {
          for (int tIndex = 0; tIndex < tStep.length; tIndex++) {
            double Vxy = gridV[iy][ix];
            cosV[tIndex][iy][ix] = cos((al - Vxy) * tStep[tIndex] / hbar);
            sinV[tIndex][iy][ix] = sin((al - Vxy) * tStep[tIndex] / hbar);
          }
        }
      }
    }
  }
}
