package physbeans.phys;

import java.awt.*;
import static java.lang.Math.*;

/**
 * Creatures color mixtures from three primary colors
 * using additive or subtractive mixing.
 * Additionally scaling factors for the primary colors can be provided.
 */
public class ThreeColorMixer extends PhysModel {
  
  protected Color primary1, primary2,primary3;
  protected boolean additive;
  protected float[] scales;
  
  protected Color mix12, mix31, mix23, mix123;
  protected Color p1s, p2s, p3s;
  
  
  /**
   * Standard constructor with additive mixture and primaries red, green, blue.
   */
  public ThreeColorMixer() {
    primary1 = Color.RED;
    primary2 = Color.GREEN;
    primary3 = Color.BLUE;
    
    additive = true;
    scales = new float[] {1.0f, 1.0f, 1.0f};
    
    computeColorMixtures();
    trigger();
  }
  
  /**
   * Gets first primary color.
   */
  public Color getFirstPrimary() {
    return primary1;
  }
  
  /**
   * Sets first primary color.
   */
  public void setFirstPrimary(Color c) {
    primary1 = c;
    computeColorMixtures();
    trigger();
  }
  
  /**
   * Gets second primary color.
   */
  public Color getSecondPrimary() {
    return primary2;
  }
  
  /**
   * Sets second primary color.
   */
  public void setSecondPrimary(Color c) {
    primary2 = c;
    computeColorMixtures();
    trigger();
  }
  
  /**
   * Gets first third color.
   */
  public Color getThirdPrimary() {
    return primary3;
  }
  
  /**
   * Sets third primary color.
   */
  public void setThirdPrimary(Color c) {
    primary3 = c;
    computeColorMixtures();
    trigger();
  }
  
  /**
   * Returns mixing method (true for additive, false for subtractive).
   */
  public boolean isAdditive() {
    return additive;
  }
  
  /**
   * Returns mixing method (true for additive, false for subtractive).
   */
  public void setAdditive(boolean b) {
    additive = b;
    computeColorMixtures();
    trigger();
  }
  
  /**
   * Returns mixture of primaries 1 and 2.
   */
  public Color getMixture12() {
    return mix12;
  }
  
  /**
   * Returns mixture of primaries 1 and 3.
   */
  public Color getMixture13() {
    return mix31;
  }
  
  /**
   * Returns mixture of primaries 2 and 3.
   */
  public Color getMixture23() {
    return mix23;
  }
  
  /**
   * Returns mixture of primaries 1, 2 and 3.
   */
  public Color getMixture123() {
    return mix123;
  }
  
  /**
   * Returns all colors (including scaled primaries).
   */
  public Color[] getScaledColors() {
    return new Color[] {p1s, p2s, p3s, mix12, mix23, mix31, mix123};
  }
  
  /**
   * Set scaling factor for a given primary.
   * For additive mixture, a small scale value means a dark color (little light),
   * for subtractive mixture a light color (little paint).
   */
  public void setScale(int i, double scale) {
    scales[i] = (float) scale;
    computeColorMixtures();
    trigger();
  }
  
  
  /**
   * Computes all color mixtures.
   */
  protected void computeColorMixtures() {
    if (additive) {
      p1s = blendColors(primary1, Color.BLACK, scales[0]);
      p2s = blendColors(primary2, Color.BLACK, scales[1]);
      p3s = blendColors(primary3, Color.BLACK, scales[2]);
      
      mix12  = mixAdditive(p1s, p2s);
      mix31  = mixAdditive(p1s, p3s);
      mix23  = mixAdditive(p2s, p3s);
      mix123 = mixAdditive(p1s, mix23);
    } else {
      p1s = blendColors(primary1, Color.WHITE, scales[0]);
      p2s = blendColors(primary2, Color.WHITE, scales[1]);
      p3s = blendColors(primary3, Color.WHITE, scales[2]);
      
      mix12  = mixSubtractive(p1s, p2s);
      mix31  = mixSubtractive(p1s, p3s);
      mix23  = mixSubtractive(p2s, p3s);
      mix123 = mixSubtractive(p1s, mix23);
    }
  }
  
  /**
   * Computes additive mixture of two colors
   */
  protected Color mixAdditive(Color c1, Color c2) {
    float[] rgb1 = c1.getRGBColorComponents(null);
    float[] rgb2 = c2.getRGBColorComponents(null);
    
    float[] rgbMix = new float[3];
    for (int i = 0; i < 3; i++) {
      rgbMix[i] = rgb1[i] + rgb2[i];
    }
    
    // correct brightness
    float b = max(getBrightness(rgb1), getBrightness(rgb2));
    float scale = b/getBrightness(rgbMix);
    for (int i = 0; i < 3; i++) {
      rgbMix[i] *= scale;
    }
    
    return new Color(rgbMix[0], rgbMix[1], rgbMix[2]);
  }
  
  /**
   * Computes subtractive mixture of two colors
   */
  protected Color mixSubtractive(Color c1, Color c2) {
    float[] rgb1 = c1.getRGBColorComponents(null);
    float[] rgb2 = c2.getRGBColorComponents(null);
    
    float[] rgbMix = new float[3];
    for (int i = 0; i < 3; i++) {
      rgbMix[i] = rgb1[i] * rgb2[i];
    }
    
    return new Color(rgbMix[0], rgbMix[1], rgbMix[2]);
  }
  
  /**
   * Computes the brightness of a color given by its (float) rgb values.
   */
  protected float getBrightness(float[] rgb) {
    // the brightness is just the maximum of r, g, b
    //    cf. java.awt.Color.RGBtoHSB
    return max(max(max(rgb[0], rgb[1]), rgb[2]), 1e-10f);
  }
  
  /**
   * Blends a color (col) into another one (colRef) according to a given scale 
   * value between 0 (full colRef) and 1 (full col).
   */
  protected Color blendColors(Color col, Color colRef, float scale) {
    scale = min(1.0f, max(0.0f, scale));
    float[] rgb = col.getRGBColorComponents(null);
    float[] rgbRef = colRef.getRGBColorComponents(null);
    for (int i = 0; i < 3; i++) {
      rgb[i] = scale*rgb[i] + (1.0f - scale)*rgbRef[i];
    }
    
    return new Color(rgb[0], rgb[1], rgb[2]);
  }
}
