package physbeans.phys;

import physbeans.func.*;
import physbeans.math.*;

import java.io.*;

/**
 * A physical bean describing  a two-dimensional harmonic oscillator with friction<br>
 *    M x'' + B x' + C x = 0<br>
 * where M is diagonal and B, C are symmetric
 * <p>
 * In addition to the four initial values it uses the following 8 parameters
 * which define the matrices M, C, B in an obvious way:
 *  <ul>
 * <li>oscillator masses m1, m2</li>
 * <li>friction coefficients b11, b12, b22</li>
 * <li>spring constants c11, c12, c22</li>
 *  </ul>
 */
public class TwoSprings extends SolveODEFunction implements Serializable {
  
  protected double m1, m2;
  protected double c11, c12, c22;
  protected double b11, b12, b22;
  
  /** Creates a new instance of DrivenSpring */
  public TwoSprings() {
    super();
    
    m1 = 1.0;
    m2 = 1.0;
    c11 = 1.0;
    c12 = 1.0;
    c22 = 1.0;
    b11 = 0.0;
    b12 = 0.0;
    b22 = 0.0;
    
    setOde(new TwoSpringsODE());
    restart();
  }
  
  /**
   * Gets mass m1.
   */
  public double getM1() {
    return this.m1;
  }
  
  /**
   * Sets mass m1.
   */
  public void setM1(double m) {
    this.m1 = m;
    restart();
  }
  
  /**
   * Gets mass m2.
   */
  public double getM2() {
    return this.m2;
  }
  
  /**
   * Sets mass m2.
   */
  public void setM2(double m) {
    this.m2 = m;
    restart();
  }
  
  /**
   * Gets stiffness coefficient c11.
   */
  public double getC11() {
    return this.c11;
  }
  
  /**
   * Sets stiffness coefficient c11.
   */
  public void setC11(double c) {
    this.c11 = c;
    restart();
  }
  
  /**
   * Gets stiffness coefficient c12.
   */
  public double getC12() {
    return this.c12;
  }
  
  /**
   * Sets stiffness coefficient c12.
   */
  public void setC12(double c) {
    this.c12 = c;
    restart();
  }
  
  /**
   * Gets stiffness coefficient c22.
   */
  public double getC22() {
    return this.c22;
  }
  
  /**
   * Sets stiffness coefficient c22.
   */
  public void setC22(double c) {
    this.c22 = c;
    restart();
  }
  
  /**
   * Gets viscous damping coefficient b11.
   */
  public double getB11() {
    return this.b11;
  }
  
  /**
   * Sets viscous damping coefficient b11.
   */
  public void setB11(double b) {
    this.b11 = b;
    restart();
  }
  
  /**
   * Gets viscous damping coefficient b12.
   */
  public double getB12() {
    return this.b12;
  }
  
  /**
   * Sets viscous damping coefficient b12.
   */
  public void setB12(double b) {
    this.b12 = b;
    restart();
  }
  
  /**
   * Gets viscous damping coefficient b22.
   */
  public double getB22() {
    return this.b22;
  }
  
  /**
   * Sets viscous damping coefficient b22.
   */
  public void setB22(double b) {
    this.b22 = b;
    restart();
  }
  
  /**
   * Gets initial positions of the two masses.
   */
  public DVector getX0() {
    return getOde().getX0().getDVector(0,1);
  }
  
  /**
   * Sets initial positions of the two masses.
   */
  public void setX0(DVector x0) {
    setInitialValue(0, x0.get(0));
    setInitialValue(1, x0.get(1));
  }
  
  /**
   * Gets initial velocities of the two masses.
   */
  public DVector getV0() {
    return getOde().getX0().getDVector(2,3);
  }
  
  /**
   * Sets initial velocities of the two masses.
   */
  public void setV0(DVector v0) {
    setInitialValue(2, v0.get(0));
    setInitialValue(3, v0.get(1));
  }
  
  class TwoSpringsODE extends ODE {
    public TwoSpringsODE() {
      super(new DVector(1.0, 0.0, 0.0, 0.0), 0.0);
    }
    
    public DVector f(DVector x, double t) {
      DVector result = new DVector(4);
      
      double x1 = x.get(0);
      double x2 = x.get(1);
      double v1 = x.get(2);
      double v2 = x.get(3);
      
      result.set(0, v1);   // x1' = v1
      result.set(1, v2);   // x2' = v2
      result.set(2, (-b11*v1 - b12*v2 - c11*x1 - c12*x2)/m1);
      result.set(3, (-b12*v1 - b22*v2 - c12*x1 - c22*x2)/m2);
   
      return result;
    }
    
  }
}
