package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.inout.ScreenWorldTrafo;
import physbeans.phys.*;

import physbeans.math.*;
import physbeans.model.*;

/**
 * View showing a LightRayBundle as a number of polylines, usually extending
 * out of the screen.
 */
public class LightRayBundleView extends View {
  
  protected LightRayBundle model;
  protected FunctionView rayView;
  protected boolean[] showRays;
  
  /** Creates a LightRayBundleView without trafo and parent. */
  public LightRayBundleView() {
    rayView = new FunctionView();
    rayView.setLineColor(new Color(0, 98, 205));
  }
  
  /**
   * Paints the light rays as polylines, including a line from the last point
   * to the border of the screen using the exit angle.
   */
  public void paintLayer(Graphics2D g2) {
    if (model == null) {
      return;  // do nothing
    }
    
    double xmax = trafo.getMaxWorld().get(0);
    double xmin = trafo.getMinWorld().get(0);
    double ymax = trafo.getMaxWorld().get(1);
    double ymin = trafo.getMinWorld().get(1);
    Point2dVector points;
    int nRays = model.getNRays();
    for (int i = 0; i < nRays; i++) {
      if (showRays != null && showRays[i] == false) {
        continue;
      }
      
      points = model.getRay(i).getPoints().copy();
      double alpha = model.getRay(i).getAngle();
      if (!Double.isNaN(alpha)) {  // compute additional point at screen border
        double xl = model.getRay(i).getLastPoint().get(0);
        double yl = model.getRay(i).getLastPoint().get(1);
        double x,y;
        if (Maths.isEqual(Math.abs(alpha), Math.PI/2)) {
          x = xl;
          y = (alpha > 0)? ymax : xmin;
        } else {
          x = (Math.abs(alpha) > Math.PI/2)? xmin : xmax;
          y = yl + (x - xl)*Math.tan(alpha);
        }
        
        points.addPoint(x, y);
      }
      
      rayView.setModel(points);
      rayView.paintLayer(g2);
    }
  }
  
  /**
   * Gets underlying model defining the light rays to paint.
   */
  public LightRayBundle getModel() {
    return model;
  }
  
  /**
   * Sets underlying model defining the light rays to paint.
   * By default all rays of the bundle are shown.
   */
  public void setModel(LightRayBundle model) {
    this.model = model;
    update();
  }
  
  /**
   * Gets color of the lines.
   */
  public Color getLineColor() {
    return rayView.getLineColor();
  }
  
  /**
   * Sets color of the lines.
   */
  public void setLineColor(Color lineColor) {
    rayView.setLineColor(lineColor);
    update();
  }
  
  /**
   * Gets visibility status of all rays.
   */
  public boolean[] getShowRays() {
    return this.showRays;
  }
  
  /**
   * Sets visibility status of all rays.
   */
  public void setShowRays(boolean[] showRays) {
    this.showRays = showRays;
    update();
  }
  
  /**
   * Sets trafo between screen and world coordinates.
   */
  public void setTrafo(ScreenWorldTrafo trafo) {
    super.setTrafo(trafo);
    rayView.setTrafo(trafo);
  }
}
