package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;

import physbeans.inout.*;
import physbeans.phys.*;
import physbeans.math.*;
import physbeans.model.*;

import static java.lang.Math.*;     // readable math at last!
import static physbeans.math.Maths.*;

/**
 * Displays a spherical mirror optionally with focal point.
 */
public class ParabolicMirrorView extends View {
  
  protected ParabolicMirror model;
  protected Color mirrorColor;
  protected boolean showFocalPoint;
  protected PointView focalView;
  
  /** Creates a new instance of SphericalMirrorView */
  public ParabolicMirrorView() {
    mirrorColor = Color.BLACK;
    showFocalPoint = false;
    focalView = new PointView();
    focalView.setPointRadius(0.2);
    focalView.setPointColor(Color.black);
  }
  
  public void paintLayer(Graphics2D g2) {
    if (model == null) {
      return;  // do nothing
    }
    
    double x = model.getPosition();
    double h = model.getHeight();
    double r = model.getRadius();
    
    double xh = h*h/(2*r);
    QuadCurve2D parabola = new QuadCurve2D.Double(xh+x, h, -xh+x, 0.0, xh+x, -h);
    
    g2.setColor(mirrorColor);
    g2.draw(parabola);
    
    if (showFocalPoint) {
      focalView.setModel(model.getFocalPoints());
      focalView.paintLayer(g2);
    }
  }
  
  /**
   * Gets underlying model defining the mirror to paint.
   */
  public ParabolicMirror getModel() {
    return model;
  }
  
  /**
   * Sets underlying model defining the mirror to paint.
   */
  public void setModel(ParabolicMirror model) {
    this.model = model;
    update();
  }
  
  /**
   * Gets the line color of the mirror.
   */
  public Color getMirrorColor() {
    return this.mirrorColor;
  }
  
  /**
   * Sets the line color of the mirror.
   */
  public void setMirrorColor(Color mirrorColor) {
    this.mirrorColor = mirrorColor;
    update();
  }
  
  /**
   * Gets the value of the showFocalPoint flag.
   */
  public boolean isShowFocalPoint() {
    return this.showFocalPoint;
  }
  
  /**
   * Sets the value of the showFocalPoint flag.
   */
  public void setShowFocalPoint(boolean showFocalPoint) {
    this.showFocalPoint = showFocalPoint;
    update();
  }
  
  /**
   * Gets radius of the circle representing the focal point.
   */
  public double getFocalPointRadius() {
    return focalView.getPointRadius();
  }
  
  /**
   * Sets radius of the circle representing the focal point.
   */
  public void setFocalPointRadius(double r) {
    focalView.setPointRadius(r);
    update();
  }
  
  /**
   * Sets trafo between screen and world coordinates.
   */
  public void setTrafo(ScreenWorldTrafo trafo) {
    super.setTrafo(trafo);
    focalView.setTrafo(trafo);
  }
  
}
