package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.inout.*;
import physbeans.math.*;

/**
 * Displays a swinging pendulum, optionally with additionally vector.
 */
public class PendulumView extends RestartableView implements Serializable {
  
  protected SingleMassView mass;
  protected SingleVectorView force;
  
  protected double length;
  protected boolean showArrow;
  protected DVector center;
  
  protected DVector mPos;
  
  /**
   * Creates a standard PendulumView.
   */
  public PendulumView() {
    super();
    
    model = new DVector(0.5, 0.5);
    initialValues = new DVector(0.5, 0.5);
    center = new DVector(2);
    
    mass = new SingleMassView();
    force = new SingleVectorView();
    
    mass.setSquare(false);
    
    length = 0.6;
    showArrow = false;
    setBallRadius(0.07);
    mPos = new DVector(2);
  }
  
  /**
   * redraw the complete pendulum according to driving parameters
   */
  public void paintLayer(Graphics2D g) {
    
    double phi = model.get(0);
    // draw line from fix point to mass
    Line2D.Double line = new Line2D.Double();
    mPos.set(0, center.get(0) + length*Math.sin(phi));
    mPos.set(1, center.get(1) - length*Math.cos(phi));
    line.setLine(center.x(), center.y(), mPos.x(), mPos.y());
    g.draw(line);
    
    // draw mass
    mass.setPosition(mPos);
    mass.paintLayer(g);
    
    // draw tangential arrow starting at mass center
    if (showArrow) {
      force.setPosition(mPos);
      force.setAngle(phi);
      force.setValue(model.get(1));
      force.paintLayer(g);
    }
  }
  
  /**
   * Sets angle of pendulum.
   * Conveniance function to simplify routing.
   */
  public void setAngle(double phi) {
    model.set(0, phi);
    update();
  }
  
  /**
   * Sets force of pendulum arrow.
   * Conveniance function to simplify routing.
   */
  public void setForce(double f) {
    model.set(1, f);
    update();
  }
  
  /**
   * gets length of string relative to panel size
   */
  public double getLength() {
    return this.length;
  }
  
  /**
   * sets length of string relative to panel size
   */
  public void setLength(double length) {
    this.length = length;
    update();
  }
  
  /**
   * Gets radius of pendulum mass in pixel.
   */
  public double getBallRadius() {
    return mass.getMassHeight()/2;
  }
  
  /**
   * Sets radius of pendulum mass in pixel.
   */
  public void setBallRadius(double ballRadius) {
    mass.setMassHeight(2*ballRadius);
    mass.setMassWidth(2*ballRadius);
  }
  
  /**
   * Gets position of fix point in world coordinates.
   */
  public DVector getCenter() {
    return this.center;
  }
  
  /**
   * Sets position of fix point in world coordinates.
   */
  public void setCenter(DVector center) {
    this.center = center;
    update();
  }
  
  /**
   * gets the color of the force arrow
   */
  public Color getArrowColor() {
    return force.getArrowColor();
  }
  
  /**
   * sets the color of the force arrow
   */
  public void setArrowColor(Color arrowColor) {
    force.setArrowColor(arrowColor);
  }
  
  /** Getter for property forceScale.
   * @return Value of property forceScale.
   */
  public double getArrowScale() {
    return force.getArrowScale();
  }
  
  /** Setter for property forceScale.
   * @param arrowScale New value of property forceScale.
   */
  public void setArrowScale(double arrowScale) {
    force.setArrowScale(arrowScale);
  }
  
  /**
   * gets length of arrow tip relative to arrow length
   */
  public double getArrowTipLength() {
    return force.getArrowTipLength();
  }
  
  /**
   * sets length of arrow tip relative to arrow length
   */
  public void setArrowTipLength(double arrowTipLength) {
    force.setArrowTipLength(arrowTipLength);
  }
  
  /**
   * gets width of arrow tip relative to arrow length
   */
  public double getArrowTipWidth() {
    return force.getArrowTipWidth();
  }
  
  /**
   * sets width of arrow tip relative to arrow length
   */
  public void setArrowTipWidth(double arrowTipWidth) {
    force.setArrowTipWidth(arrowTipWidth);
  }
  
  /** Getter for property showArrow.
   * @return Value of property showArrow.
   */
  public boolean isShowArrow() {
    return this.showArrow;
  }
  
  /** Setter for property showArrow.
   * @param showArrow New value of property showArrow.
   */
  public void setShowArrow(boolean showArrow) {
    this.showArrow = showArrow;
    update();
  }
  
  /**
   * used by LayeredScreen
   */
  public void setTrafo(ScreenWorldTrafo trafo) {
    trafo.setConstrained(true);
    super.setTrafo(trafo);
    mass.setTrafo(trafo);
    force.setTrafo(trafo);
  }
}
