package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.math.*;

/**
 * View that draws one of a set of simple figures that are given by specifying
 * up to four Points P1 .. P4:
 *  <table>
 *    <th><td>object</td><td>used points</td></th>
 *    <tr><td>one line</td><td>P1, P2</td></tr>
 *    <tr><td>two unconnected lines</td><td>P1, P2, P3, P4</td></tr>
 *    <tr><td>three connected lines</td><td>P1, P2, P3, P4</td></tr>
 *    <tr><td>a triangle</td><td>P1, P2, P3</td></tr>
 *    <tr><td>a quadrangle</td><td>P1, P2, P3, P4</td></tr>
 *    <tr><td>a rectangle</td><td>P1, P2</td></tr>
 *    <tr><td>an ellipse</td><td>P1, P2</td></tr>
 *    <tr><td>an arc</td><td>P1, P2</td></tr>
 *    <tr><td>a string</td><td>P1, P2</td></tr>
 *  </table>
 * Symbols can be filled (if meaningful). The position and size of the string is
 * given by the bounding rectangle defined by points P1 and P2, in the same way
 * as for the rectangle, ellipse and arc. The ellipse is soecified by two points,
 * which either define the bounding rectangle or the center and two radii.
 * The angle allows to draw the figure around the midpoint of P1/P2.
 */
public class SimpleFigureView extends View implements Serializable {

  // implemented figures
  public static final int LINE = 0;
  public static final int TWOLINES = 1;
  public static final int THREELINES = 2;
  public static final int TRIANGLE = 3;
  public static final int QUADRANGLE = 4;
  public static final int RECTANGLE = 5;
  public static final int OVAL = 6;
  public static final int CENTEREDOVAL = 7;
  public static final int ARC = 8;
  public static final int STRING = 9;
  protected int figureType;
  protected Color color;
  protected boolean filled;
  protected DVector p1;
  protected DVector p2;
  protected DVector p3;
  protected DVector p4;
  protected String string;
  protected double angle;
  protected int arcAngle;

  /** Creates a new instance of SymbolPainter */
  public SimpleFigureView() {
    super();
    figureType = QUADRANGLE;
    color = Color.black;
    filled = false;
    p1 = new DVector(0.0, 0.0);
    p2 = new DVector(0.0, 1.0);
    p3 = new DVector(1.0, 0.0);
    p4 = new DVector(1.0, 1.0);
    string = "juhu";
    angle = 0.0;
    arcAngle = 90;
  }

  /**
   * paint one of the figures
   */
  public void paintLayer(Graphics2D g2) {
    g2.setColor(color);
    g2.rotate(angle, (p2.x() + p1.x()) / 2.0, (p2.y() + p1.y()) / 2.0);

    switch (figureType) {
      case LINE:
        drawLine(g2);
        break;
      case TWOLINES:
        drawTwoLines(g2);
        break;
      case THREELINES:
        drawThreeLines(g2);
        break;
      case TRIANGLE:
        drawTriangle(g2);
        break;
      case QUADRANGLE:
        drawQuadrangle(g2);
        break;
      case RECTANGLE:
        drawRectangle(g2);
        break;
      case OVAL:
        drawOval(g2);
        break;
      case ARC:
        drawArc(g2);
        break;
      case STRING:
        drawString(g2);
        break;
      case CENTEREDOVAL:
        drawCenteredOval(g2);
        break;
    }
    g2.rotate(-angle, (p2.x() + p1.x()) / 2.0, (p2.y() + p1.y()) / 2.0);

  }

  /**
   * Gets type of figure.
   */
  public int getFigureType() {
    return this.figureType;
  }

  /**
   * Sets type of figure.
   */
  public void setFigureType(int figureType) {
    this.figureType = figureType;
    update();
  }

  /**
   * Gets color to draw/fill figure.
   */
  public Color getColor() {
    return this.color;
  }

  /**
   * Sets color to draw/fill figure.
   */
  public void setColor(Color color) {
    this.color = color;
    update();
  }

  /**
   * Gets filled flag.<br>
   * If true, figure is filled.
   */
  public boolean isFilled() {
    return this.filled;
  }

  /**
   * Sets filled flag.<br>
   * If true, figure is filled.
   */
  public void setFilled(boolean filled) {
    this.filled = filled;
    update();
  }

  /**
   * Gets point 1.
   */
  public DVector getP1() {
    return this.p1;
  }

  /**
   * Sets point 1.
   */
  public void setP1(DVector p1) {
    this.p1 = p1;
    update();
  }

  /**
   * Gets point 2.
   */
  public DVector getP2() {
    return this.p2;
  }

  /**
   * Sets point 2.
   */
  public void setP2(DVector p2) {
    this.p2 = p2;
    update();
  }

  /**
   * Gets point 3.
   */
  public DVector getP3() {
    return this.p3;
  }

  /**
   * Sets point 3.
   */
  public void setP3(DVector p3) {
    this.p3 = p3;
    update();
  }

  /**
   * Gets point 4.
   */
  public DVector getP4() {
    return this.p4;
  }

  /**
   * Sets point 4.
   */
  public void setP4(DVector p4) {
    this.p4 = p4;
    update();
  }

  /**
   * Gets string to draw.
   */
  public String getString() {
    return this.string;
  }

  /**
   * Sets string to draw.
   */
  public void setString(String string) {
    this.string = string;
    update();
  }

  /**
   * Gets angle (in rad) for rotation around the midpoint of P1/P2.
   */
  public double getAngle() {
    return angle;
  }

  /**
   * Sets angle (in rad) for rotation around the midpoint of P1/P2.
   */
  public void setAngle(double angle) {
    this.angle = angle;
    update();
  }

  /**
   * Gets arc angle of the arc.
   */
  public int getArcAngle() {
    return this.arcAngle;
  }

  /**
   * Sets arc angle of the arc.
   */
  public void setArcAngle(int arcAngle) {
    this.arcAngle = arcAngle;
    update();
  }

  protected void drawLine(Graphics2D g) {
    Line2D l = new Line2D.Double(p1.get(0), p1.get(1), p2.get(0), p2.get(1));
    g.draw(l);
  }

  protected void drawTwoLines(Graphics2D g) {
    Line2D l1 = new Line2D.Double(p1.get(0), p1.get(1), p2.get(0), p2.get(1));
    Line2D l2 = new Line2D.Double(p3.get(0), p3.get(1), p4.get(0), p4.get(1));
    g.draw(l1);
    g.draw(l2);
  }

  protected void drawThreeLines(Graphics2D g) {
    GeneralPath pline = new GeneralPath();
    pline.moveTo((float) p1.get(0), (float) p1.get(1));
    pline.lineTo((float) p2.get(0), (float) p2.get(1));
    pline.lineTo((float) p3.get(0), (float) p3.get(1));
    pline.lineTo((float) p4.get(0), (float) p4.get(1));
    g.draw(pline);
  }

  protected void drawTriangle(Graphics2D g) {
    GeneralPath pline = new GeneralPath();
    pline.moveTo((float) p1.get(0), (float) p1.get(1));
    pline.lineTo((float) p2.get(0), (float) p2.get(1));
    pline.lineTo((float) p3.get(0), (float) p3.get(1));
    pline.closePath();

    if (filled) {
      g.fill(pline);
    } else {
      g.draw(pline);
    }
  }

  protected void drawQuadrangle(Graphics2D g) {
    GeneralPath pline = new GeneralPath();
    pline.moveTo((float) p1.get(0), (float) p1.get(1));
    pline.lineTo((float) p2.get(0), (float) p2.get(1));
    pline.lineTo((float) p3.get(0), (float) p3.get(1));
    pline.lineTo((float) p4.get(0), (float) p4.get(1));
    pline.closePath();

    if (filled) {
      g.fill(pline);
    } else {
      g.draw(pline);
    }
  }

  protected void drawRectangle(Graphics2D g) {
    Rectangle2D rect = new Rectangle2D.Double(p1.get(0), p1.get(1),
            Math.abs(p2.get(0) - p1.get(0)), Math.abs(p2.get(1) - p1.get(1)));

    if (filled) {
      g.fill(rect);
    } else {
      g.draw(rect);
    }
  }

  protected void drawArc(Graphics2D g) {
    // careful with the signs: I'm using a mirrored coordinate system
    //  (namely: x to the right, y upwards)!
    Arc2D arc = new Arc2D.Double(p1.get(0), p1.get(1),
            Math.abs(p2.get(0) - p1.get(0)), Math.abs(p2.get(1) - p1.get(1)),
            0.0, -arcAngle, Arc2D.OPEN);

    if (filled) {
      g.fill(arc);
    } else {
      g.draw(arc);
    }
  }

  protected void drawOval(Graphics2D g) {
    Ellipse2D elli = new Ellipse2D.Double(p1.get(0), p1.get(1),
            Math.abs(p2.get(0) - p1.get(0)), Math.abs(p2.get(1) - p1.get(1)));

    if (filled) {
      g.fill(elli);
    } else {
      g.draw(elli);
    }
  }

  protected void drawCenteredOval(Graphics2D g) {
    Ellipse2D elli = new Ellipse2D.Double();
    elli.setFrameFromCenter(p1.get(0), p1.get(1),
            p1.get(0) + p2.get(0), p1.get(1) + p2.get(1));

    if (filled) {
      g.fill(elli);
    } else {
      g.draw(elli);
    }
  }

  protected void drawString(Graphics2D g) {
    AffineTransform oldTrafo = g.getTransform();

    Rectangle2D bounds = g.getFontMetrics().getStringBounds(string, g);
    double sx = Math.abs(p2.get(0) - p1.get(0)) / bounds.getWidth();
    double sy = Math.abs(p2.get(1) - p1.get(1)) / bounds.getHeight();
    g.translate(p1.get(0), p1.get(1));
    g.scale(sx, -sy);
    g.drawString(string, 0.0f, 0.0f);
    g.setTransform(oldTrafo);
  }
}
