package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import physbeans.inout.*;
import physbeans.math.*;
import physbeans.phys.*;
import physbeans.event.*;

/**
 * View that displays a LinearChainOscillator showing circles or squares for
 * the masses and zigzag lines for the springs. Additionally external forces
 * acting on the masses can be displayed as red arrows.
 * The masses are lined up horizontally and move horizontally or vertically,
 * according to the flag transversal.
 * The size of the mass symbols and the width of the spring symbols can be
 * adapted to the masses resp. spring constants.
 */
public class SpringChainView extends View {

  protected DVector x,  f,  x0,  f0;
  protected double scaleX,  scaleF;
  protected double xBase,  xBase0,  scaleXBase,  basementWidth;
  protected boolean square,  hideHugeSprings,  showArrows,  transversal,  coiled;
  protected LinearChainOscillator system;
  protected double massScale,  massDefaultSize;
  protected double springScale,  springDefaultSize;
  protected double springDefaultLineLength;
  protected SingleMassView[] massV;
  protected SingleVectorView[] forceV;
  protected SingleSpringView[] springV;
  protected final double THIRD = 0.3333333333333333;

  /**
   * Creates a new instance of SpringChainView.
   */
  public SpringChainView() {
    super();

    scaleX = 0.2;
    scaleF = 0.5;
    scaleXBase = 1.0;
    xBase = 0.0;
    xBase0 = 0.0;

    square = true;
    hideHugeSprings = false;
    showArrows = false;
    transversal = false;
    coiled = true;
    system = null;
    massScale = 1.0;
    massDefaultSize = 0.4;
    springScale = 1.0;
    springDefaultSize = 0.6;
    springDefaultLineLength = 0.1;
    basementWidth = 0.2;

    createViews(3);
  }

  /**
   * Redraws the complete spring according to driving parameter.
   */
  public void paintLayer(Graphics2D g) {

    int n = massV.length;
    for (int i = 0; i < n; i++) {
      DVector posi;
      if (transversal) {
        double yi = scaleX * x.get(i);
        posi = new DVector(i + 1, yi);
      } else {
        double xi = (i + 1) + scaleX * x.get(i);
        posi = new DVector(xi, 0.0);
      }

      // base point position
      double basePos = xBase * scaleXBase;
      springV[0].setLeftPosition(new DVector(basePos, 0.0));
      paintBasement(g, basePos);

      massV[i].setPosition(posi);
      springV[i].setRightPosition(posi);
      springV[i + 1].setLeftPosition(posi);
    }

    // paint springs then masses
    for (int i = 0; i < n; i++) {
      springV[i].paintLayer(g);
    }
    springV[n].paintLayer(g);
    for (int i = 0; i < n; i++) {
      massV[i].paintLayer(g);
    }

    if (showArrows) {
      for (int i = 0; i < n; i++) {
        double xi = (i + 1) + scaleX * x.get(i);
        DVector posi = new DVector(xi, 0.0);
        forceV[i].setPosition(posi);
        forceV[i].setValue(f.get(i));
        forceV[i].paintLayer(g);
      }
    }
  }

  /**
   * Gets LinearChainOscillator used to get masses, springs etc. to adapt sizes,
   * number of spring turns etc. automatically.
   * <P>If it is not set, all subsystems have the same dimensions.
   */
  public LinearChainOscillator getSystem() {
    return system;
  }

  /**
   * Sets LinearChainOscillator used to get masses, springs etc. to adapt sizes,
   * number of spring turns etc. automatically.
   * <P>If it is not set, all subsystems have the same dimensions.
   */
  public void setSystem(LinearChainOscillator s) {
    system = s;

    if (system != null) {
      // get current parameters from system
      checkLength(system.getMasses());
      setDrawingParameters();
      update();

      // and stay informed of future changes
      system.addTriggerListener(new TriggerListener() {

        public void sourceChanged(TriggerEvent e) {
          setDrawingParameters();
          update();
        }
      });
    }
  }

  /**
   * Sets the displacements of the masses.
   */
  public void setDisplacements(DVector xNew) {
    checkLength(xNew);
    x = xNew;
    update();
  }

  /**
   * Gets the displacements of the masses.
   */
  public DVector getDisplacements() {
    return x;
  }

  /**
   * Sets the forces acting on the masses.
   */
  public void setForces(DVector fNew) {
    checkLength(fNew);
    f = fNew;
    update();
  }

  /**
   * Gets the forces acting on the masses.
   */
  public DVector getForces() {
    return f;
  }

  /**
   * Sets the (x-)position of the base point.
   */
  public void setBase(double xNew) {
    xBase = xNew;
    update();
  }

  /**
   * Gets the (x-)position of the base point.
   */
  public double getBase() {
    return xBase;
  }

  /**
   * Sets the scale factor of the base point position.
   */
  public void setBaseScale(double xNew) {
    scaleXBase = xNew;
    update();
  }

  /**
   * Gets the scale factor of the base point position.
   */
  public double getBaseScale() {
    return scaleXBase;
  }

  /**
   * Sets the initial base point position.
   */
  public void setInitialBase(double x) {
    xBase0 = x;
    xBase = x;
    update();
  }

  /**
   * Gets the initial base point position.
   */
  public double getInitialBase() {
    return xBase0;
  }

  /**
   * Sets the width of the basement marker.
   */
  public void setBasementWidth(double b) {
    basementWidth = b;
    update();
  }

  /**
   * Gets the width of the basement marker.
   */
  public double getBasementWidth() {
    return basementWidth;
  }

  /**
   * Gets the scale factor of the x displacement.
   */
  public double getXScale() {
    return scaleX;
  }

  /**
   * Sets the scale factor of the x displacement.
   */
  public void setXScale(double s) {
    scaleX = s;
    update();
  }

  /**
   * Gets the scale factor of the force vectors.
   */
  public double getFScale() {
    return scaleF;
  }

  /**
   * Sets the scale factor of the force vectors.
   */
  public void setFScale(double f) {
    scaleF = f;
    for (int i = 0; i < x.getDimension(); i++) {
      forceV[i].setArrowScale(scaleF);
    }
    update();
  }

  /**
   * Gets the displacements of the masses.
   */
  public DVector getInitialDisplacements() {
    return x0;
  }

  /**
   * Sets the displacements of the masses.
   */
  public void setInitialDisplacements(DVector x0New) {
    checkLength(x0New);
    x0 = x0New;
    x = x0New;
    update();
  }

  /**
   * Set the displacement of mass i.
   * <P>Conveniance function to simplify message routing.
   */
  public void setInitialDisplacement(int i, double d) {
    x0.set(i, d);
    x.set(i, d);
    update();
  }

  /**
   * Gets the initial forces acting on the masses.
   */
  public DVector getInitialForces() {
    return f0;
  }

  /**
   * Sets the initial forces acting on the masses.
   */
  public void setInitialForces(DVector f0New) {
    checkLength(f0New);
    f0 = f0New;
    f = f0New;
    update();
  }

  /**
   * Set the initial force acting on mass i.
   * <P>Conveniance function to simplify message routing.
   */
  public void setInitialForce(int i, double d) {
    f0.set(i, d);
    f.set(i, d);
    update();
  }

  /**
   * Resets the model to the initial values and redraws.
   */
  public void restart() {
    x = x0.copy();
    f = f0.copy();
    xBase = xBase0;
    update();
  }

  /**
   * Transfers coordinate system from LayeredScreen to all subviews.
   */
  public void setTrafo(ScreenWorldTrafo trafo) {
    super.setTrafo(trafo);
    int n = massV.length;
    for (int i = 0; i < n; i++) {
      springV[i].setTrafo(trafo);
      massV[i].setTrafo(trafo);
    }
    springV[n].setTrafo(trafo);
  }

  /**
   * Gets the scale factor defining the size of the masses.
   */
  public double getMassScale() {
    return massScale;
  }

  /**
   * Sets the scale factor defining the size of the masses.
   */
  public void setMassScale(double ms) {
    massScale = ms;
    setDrawingParameters();
    update();
  }

  /**
   * Gets the scale factor defining the size of the springs.
   */
  public double getSpringScale() {
    return springScale;
  }

  /**
   * Sets the scale factor defining the size of the springs.
   */
  public void setSpringScale(double ss) {
    springScale = ss;
    setDrawingParameters();
    update();
  }

  /**
   * Gets square flag.
   * <P>If true, all masses are shown as rectangle, else as ovals.
   */
  public boolean isSquare() {
    return square;
  }

  /**
   * Sets square flag.
   * <P>If true, all masses are shown as rectangle, else as ovals.
   */
  public void setSquare(boolean b) {
    square = b;
    setDrawingParameters();
    update();
  }

  /**
   * Gets hideHugeSprings flag.
   * <P>If true, very large springs (corresponding to very low stiffness)
   * are not drawn at all.
   */
  public boolean isHideHugeSprings() {
    return hideHugeSprings;
  }

  /**
   * Sets hideHugeSprings flag.
   * <P>If true, very large springs (corresponding to very low stiffness)
   * are not drawn at all.
   */
  public void setHideHugeSprings(boolean b) {
    hideHugeSprings = b;
    setDrawingParameters();
    update();
  }

  /**
   * Gets showArrows flag.
   *<P>If true, external forces are shown as red arrows attached to the masses.
   */
  public boolean isShowArrows() {
    return showArrows;
  }

  /**
   * Sets showArrows flag.
   *<P>If true, external forces are shown as red arrows attached to the masses.
   */
  public void setShowArrows(boolean b) {
    showArrows = b;
    update();
  }

  /**
   * Gets transversal flag.
   *<P>If true, the masses move vertically, else horizontally.
   */
  public boolean isTransversal() {
    return transversal;
  }

  /**
   * Sets transversal flag.
   *<P>If true, the masses move vertically, else horizontally.
   */
  public void setTransversal(boolean b) {
    transversal = b;
    update();
  }

  /**
   * Gets coiled flag.
   *<P>If true, a symbolic coil is drawn, else a simple rubber band.
   */
  public boolean isCoiled() {
    return coiled;
  }

  /**
   * Sets coiled flag.
   *<P>If true, a symbolic coil is drawn, else a simple rubber band.
   */
  public void setCoiled(boolean b) {
    coiled = b;
    setDrawingParameters();
    update();
  }

  /**
   * Creates internal structures necessary for n masses.
   */
  protected void createViews(int n) {
    x = new DVector(n);
    f = new DVector(n);
    x0 = new DVector(n);
    f0 = new DVector(n);

    massV = new SingleMassView[n];
    forceV = new SingleVectorView[n];
    springV = new SingleSpringView[n + 1];

    for (int i = 0; i < n; i++) {
      massV[i] = new SingleMassView();
      forceV[i] = new SingleVectorView();
      springV[i] = new SingleSpringView();
    }
    springV[n] = new SingleSpringView();

    springV[0].setLeftPosition(new DVector(xBase * scaleXBase, 0.0));
    springV[n].setRightPosition(new DVector(n + 1, 0.0));

    setDrawingParameters();
  }

  /**
   * Sets all parameters that are used for drawing, i.e. mass width and height,
   * width of the spring,...
   */
  protected void setDrawingParameters() {
    int n = x.getDimension();
    double mw = massScale * massDefaultSize;
    for (int i = 0; i < n; i++) {
      double m = mw;
      if (system != null) {
        m *= Math.pow(system.getMasses().get(i), THIRD);
      }
      massV[i].setMassHeight(m);
      massV[i].setMassWidth(m);
      massV[i].setSquare(square);

      forceV[i].setArrowScale(scaleF);
    }

    // adapt spring appearance
    double sw = springScale * springDefaultSize;
    for (int i = 0; i < n + 1; i++) {
      springV[i].setCoiled(coiled);

      double s = sw;
      if (system != null) {
        s /= Math.pow(system.getSpringStiffnesses().get(i), THIRD);
      }
      if (hideHugeSprings && Maths.isZero(1.0 / s)) {
        springV[i].setSpringWidth(-1.0);  // hide spring
      } else {
        springV[i].setSpringWidth(s);
      }
    }

    // extend spring end lines to fit mass width
    for (int i = 0; i < n; i++) {
      double extralong = springDefaultLineLength + massV[i].getMassWidth() / 2;
      springV[i].setEndLineLength(extralong);
      springV[i + 1].setStartLineLength(extralong);
    }
  }

  /**
   * Adapts view to size of input vector, if necessary.
   */
  protected void checkLength(DVector v) {
    if (v.getDimension() != massV.length) {
      createViews(v.getDimension());
    }
  }

  /**
   * Draws a rectangle for the basement at basePos.
   */
  protected void paintBasement(Graphics2D g, double basePos) {
    Line2D l = new Line2D.Double(basePos, -basementWidth / 2, basePos, basementWidth / 2);
    g.draw(l);
  }
}
