package physbeans.views;

import java.awt.*;
import java.io.*;

import physbeans.event.*;
import physbeans.inout.ScreenWorldTrafo;

/**
 * Base class of classes that provide a graphical representation of a
 * views-specific model.
 * <P>A View can paint onto a screen using world coordinates from an associated
 * LayeredScreen, its parent. A view has a depth, i.e. a position in z
 * direction increasing with the distance from the viewer.
 * <P>Views send trigger events to inform their listeners (usually a
 * LayeredScreen) of changes.
 */
abstract public class View implements Serializable, Comparable<View> {
  
  protected ScreenWorldTrafo trafo;
  protected Container myParent;
  protected int depth;
  
  protected transient TriggerSupport trigSup;
  
  /**
   * default constructor
   */
  public View() {
    trafo = new ScreenWorldTrafo();
    trigSup = new TriggerSupport();
    depth = 50;
  }
  
  /**
   * Do all initialisations that can only be done after the layout (ie the size)
   * is fixed.
   */
  public void doLayout() {
    // default: do nothing
    // a view that needs to do something at layout time has
    //   to overwrite this function
  }
  
  /*
   * update view<br>
   * asks container (a LayeredScreen) to repaint
   */
  public void update() {
    if (myParent != null) {
      myParent.repaint();
    }
  }
  
  abstract public void paintLayer(Graphics2D g2);
  
  /**
   * sets containing element<br>
   * used by LayeredScreen
   */
  public void setParent(Container c) {
    myParent = c;
  }
  
  /**
   * sets new trafo<br>
   * used by LayeredScreen
   */
  public void setTrafo(ScreenWorldTrafo trafo) {
    this.trafo = trafo;
  }
  
  /**
   * Gets the depth of the view, i.e. its position in z direction increasing
   * with the distance from the viewer.
   */
  public int getDepth() {
    return depth;
  }
  
  /**
   * Sets the depth of the view, i.e. its position in z direction increasing
   * with the distance from the viewer.
   * <P>A LayeredScreen draws its views in order of descending depth so that
   * opaque portions of a view with smaller depth obscure views with larger
   * depths.
   */
  public void setDepth(int d) {
    depth = d;
  }
  
  /**
   * Add a TriggerListener.
   * <P>Triggers immediately to announce the view (usually to a screen).
   */
  public synchronized void addTriggerListener(TriggerListener l) {
    trigSup.addTriggerListener(l);
    trigger();   // inform listener immediately
  }
  
  /**
   * Remove a TriggerListener.
   */
  public synchronized void removeTriggerListener(TriggerListener l) {
    trigSup.removeTriggerListener(l);
  }
  
  /**
   * Fire a Trigger event.
   */
   protected void fireTriggerEvent(TriggerEvent e) {
    trigSup.fireTriggerEvent(e);
  }
  
  /**
   * Informs listeners of global changes.
   */
  protected void trigger() {
    fireTriggerEvent(new TriggerEvent(this));
  }
  
  /**
   * Uses the depth along the visual (z) direction to compare views.
   * Returns 1 (0, -1), if this view has a smaller (equal, larger) depth
   * than the specified object o. This leads to a reverse ordering, i.e. views
   * with larger depth come first.
   */
  public int compareTo(View v) {
    if (this.getDepth() < v.getDepth()) {
      return 1;
    } else if (this.getDepth() > v.getDepth()) {
      return -1;
    } else {
      return 0;
    }
  }
}
